r"""
The downanlysis of cellphonedb
"""

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.patches as mpatches
import scanpy as sc
import matplotlib
import anndata

kpy_install=False

def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)

def check_kpy():
    """
    
    """
    global kpy_install
    try:
        import ktplotspy as kpy
        kpy_install=True
        print('ktplotspy have been install version:',kpy.__version__)
    except ImportError:
        raise ImportError(
            'Please install the ktplotspy: `pip install ktplotspy`.'
        )

def cpdb_network_cal(adata:anndata.AnnData,pvals:list,celltype_key:str)->dict:
    r"""
    Calculate a CPDB (Cell Phone Database) network using gene expression data and return a dictionary of results.

    Arguments:
        adata: An AnnData object containing gene expression data.
        pvals: A list or array of p-values for each gene in the expression data.
        celltype_key: The name of the cell type key in adata.obs.

    Returns:
        cpdb_dict: A dictionary of results generated by `kpy.plot_cpdb_heatmap()` from the `ktplotspy` library.

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    cpdb_dict=kpy.plot_cpdb_heatmap(
        adata = adata,
        pvals = pvals,
        celltype_key = celltype_key,
        figsize = (1,1),
        title = "",
        symmetrical = False,
        return_tables=True,
    )
    return cpdb_dict


def cpdb_plot_network(adata:anndata.AnnData,interaction_edges:pd.DataFrame,
                      celltype_key:str,nodecolor_dict=None,
                      edgeswidth_scale:int=10,nodesize_scale:int=1,
                      pos_scale:int=1,pos_size:int=10,figsize:tuple=(5,5),title:str='',
                      legend_ncol:int=3,legend_bbox:tuple=(1,0.2),legend_fontsize:int=10,
                     return_graph:bool=False)->nx.Graph:
    r"""
    Plot a network of interactions between cell types using gene expression data.

    Arguments:
        adata: An AnnData object containing gene expression data.
        interaction_edges: A DataFrame containing the edges of the network, including the source, target, and count of interactions.
        celltype_key: The name of the cell type key in adata.obs.
        nodecolor_dict: A dictionary mapping cell type names to node colors. If not provided, uses the colors specified in adata.uns.
        edgeswidth_scale: The scaling factor for edge width. Default is 10.
        nodesize_scale: The scaling factor for node size. Default is 1.
        pos_scale: The scaling factor for node positions. Default is 1.
        pos_size: The size of the node positions. Default is 10.
        figsize: The size of the plot. Default is (5,5).
        title: The title of the plot. Default is ''.
        legend_ncol: The number of columns in the legend. Default is 3.
        legend_bbox: The location of the legend. Default is (1,0.2).
        legend_fontsize: The font size of the legend. Default is 10.
        return_graph: If True, returns the network graph. Default is False.

    Returns:
        ax or G: The plot if `return_graph` is False, otherwise the network graph.

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    #set Digraph of cellphonedb
    G=nx.DiGraph()
    for i in interaction_edges.index:
        G.add_edge(interaction_edges.loc[i,'SOURCE'],
                   interaction_edges.loc[i,'TARGET'],
                   weight=interaction_edges.loc[i,'COUNT'],)
    
    #set celltypekey's color
    if nodecolor_dict!=None:
        type_color_all=nodecolor_dict
    else:
        if '{}_colors'.format(celltype_key) in adata.uns:
            type_color_all=dict(zip(adata.obs[celltype_key].cat.categories,adata.uns['{}_colors'.format(celltype_key)]))
        else:
            if len(adata.obs[celltype_key].cat.categories)>28:
                type_color_all=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.default_102))
            else:
                type_color_all=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.zeileis_28))
    
    #set G_nodes_dict
    nodes=[]
    G_degree=dict(G.degree(G.nodes()))


    G_nodes_dict={}
    links = []
    for i in G.edges:
        if i[0] not in G_nodes_dict.keys():
            G_nodes_dict[i[0]]=0
        if i[1] not in G_nodes_dict.keys():
            G_nodes_dict[i[1]]=0
        links.append({"source": i[0], "target": i[1]})
        weight=G.get_edge_data(i[0],i[1])['weight']
        G_nodes_dict[i[0]]+=weight
        G_nodes_dict[i[1]]+=weight
        
    #plot
    fig, ax = plt.subplots(figsize=figsize) 
    pos = nx.spring_layout(G, scale=pos_scale, k=(pos_size)/np.sqrt(G.order()))
    p=dict(G.nodes)

    nodesize=np.array([G_nodes_dict[u] for u in G.nodes()])/nodesize_scale
    nodecolos=[type_color_all[u] for u in G.nodes()]
    nx.draw_networkx_nodes(G, pos, nodelist=p,node_size=nodesize,node_color=nodecolos)

    edgewidth = np.array([G.get_edge_data(u, v)['weight'] for u, v in G.edges()])/edgeswidth_scale
    nx.draw_networkx_edges(G, pos,width=edgewidth)


    #label_options = {"ec": "white", "fc": "white", "alpha": 0.6}
    #nx.draw_networkx_labels(G, pos, font_size=10,) #bbox=label_options)
    plt.grid(False)
    plt.axis("off")
    plt.xlim(-2,2)
    plt.ylim(-2,1.5)

    labels = adata.obs[celltype_key].cat.categories
    #用label和color列表生成mpatches.Patch对象，它将作为句柄来生成legend
    color = [type_color_all[u] for u in labels]
    patches = [mpatches.Patch(color=type_color_all[u], label=u) for u in labels ] 

    #plt.xlim(-0.05, 1.05)
    #plt.ylim(-0.05, 1.05)
    plt.axis("off")
    plt.title(title)
    plt.legend(handles=patches,
               bbox_to_anchor=legend_bbox, 
               ncol=legend_ncol,
               fontsize=legend_fontsize)
    if return_graph==True:
        return G
    else:
        return ax
    #return {'Graph':G,'ax':ax}

def cpdb_plot_interaction(adata:anndata.AnnData,cell_type1:str,cell_type2:str,
                          means:pd.DataFrame,pvals:pd.DataFrame,
                          celltype_key:str,genes=None,
                         keep_significant_only:bool=True,figsize:tuple = (4,8),title:str="",
                         max_size:int=1,highlight_size:float = 0.75,standard_scale:bool = True,
                         cmap_name:str='viridis',
                         ytickslabel_fontsize:int=8,xtickslabel_fontsize:int=8,title_fontsize:int=10)->matplotlib.axes._axes.Axes:
    r"""
    Plot a CellPhoneDB interaction.

    Arguments:
        adata: AnnData object containing the data.
        cell_type1: Name of cell type 1.
        cell_type2: Name of cell type 2.
        means: DataFrame containing the means for each interaction.
        pvals: DataFrame containing the p-values for each interaction.
        celltype_key: Key in adata.obs that contains the cell type information.
        genes: List of genes to include in the plot. If None, all genes are included.
        keep_significant_only: Whether to keep only significant interactions.
        figsize: Figure size.
        title: Title of the plot.
        max_size: Maximum size of the dots.
        highlight_size: Size of the dots for the highlighted interaction.
        standard_scale: Whether to standard scale the data.
        cmap_name: Name of the colormap to use.
        ytickslabel_fontsize: Fontsize of the yticks labels.
        xtickslabel_fontsize: Fontsize of the xticks labels.
        title_fontsize: Fontsize of the title.
    
    Returns:
        ax: Axes object containing the plot.

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    fig=kpy.plot_cpdb(
        adata = adata,
        cell_type1 = cell_type1,
        cell_type2 = cell_type2, 
        means = means,
        pvals = pvals,
        celltype_key = celltype_key,
        genes = genes,
        keep_significant_only=keep_significant_only,
        figsize = figsize,
        title = "",
        max_size = max_size,
        highlight_size = highlight_size,
        standard_scale = standard_scale,
        cmap_name=cmap_name
    ).draw()
    
    #ytickslabels
    labels=fig.get_axes()[0].yaxis.get_ticklabels()
    plt.setp(labels, fontsize=ytickslabel_fontsize)

    #xtickslabels
    labels=fig.get_axes()[0].xaxis.get_ticklabels()
    plt.setp(labels, fontsize=xtickslabel_fontsize)

    fig.get_axes()[0].set_title(title,fontsize=title_fontsize)
    
    return fig.get_axes()[0]

def cpdb_submeans_exacted(means:pd.DataFrame,cell_names:str,cell_type:str='ligand')->pd.DataFrame:
    r"""
    Returns a subset of the means DataFrame for a given cell type and cell name.

    Arguments:
        means: DataFrame containing the means for each interaction.
        cell_names: Name of the cell type.
        cell_type: Whether the cell type is a ligand or a receptor.

    Returns:
        means: Subset of the means DataFrame for a given cell type and cell name.

    """
    if cell_type=='ligand':
        means_columns=means.columns[:11].tolist()+means.columns[means.columns.str.contains('{}\|'.format(cell_names))].tolist()
    elif cell_type=='receptor':
        means_columns=means.columns[:11].tolist()+means.columns[means.columns.str.contains('\|{}'.format(cell_names))].tolist()
    else:
        raise ValueError('cell_type must be ligand or receptor')
    return means.loc[:,means_columns]

def cpdb_interaction_filtered(adata:anndata.AnnData,cell_type1:str,cell_type2:str,
                              means:pd.DataFrame,pvals:pd.DataFrame,celltype_key:str,genes=None,
                         keep_significant_only:bool=True,figsize:tuple = (0,0),
                         max_size:int=1,highlight_size:float = 0.75,standard_scale:bool = True,cmap_name:str='viridis',)->list:
    r"""
    Returns a list of unique interaction groups from a CellPhoneDB analysis filtered by significance.

    Arguments:
        adata: Annotated data matrix containing normalized gene expression data.
        cell_type1: The cell type name for the first interacting partner.
        cell_type2: The cell type name for the second interacting partner.
        means: Dataframe containing the means of gene expression for each interaction.
        pvals: Dataframe containing the p-values for each interaction.
        celltype_key: The column name in the `adata.obs` dataframe containing cell type annotations.
        genes: List of gene names to filter on. If None, no filtering is applied.
        keep_significant_only: Whether to filter on significant interactions only (i.e. interactions with FDR < 0.05).
        figsize: Figure size. Width and height of the figure in inches. 
        max_size: Maximum size of the markers in the plot.
        highlight_size: Size of the markers for the highlighted interaction.
        standard_scale: Whether to standard scale the data.
        cmap_name: Name of the colormap to use.

    Returns:
        genes: List of unique interaction groups from a CellPhoneDB analysis filtered by significance.

    """
    check_kpy()
    global kpy_install
    if kpy_install==True:
        global_imports("ktplotspy","kpy")

    res=kpy.plot_cpdb(
        adata = adata,
        cell_type1 = cell_type1,
        cell_type2 = cell_type2, 
        means = means,
        pvals = pvals,
        celltype_key = celltype_key,
        genes = genes,
        keep_significant_only=keep_significant_only,
        figsize = figsize,
        title = "",
        max_size = max_size,
        highlight_size = highlight_size,
        standard_scale = standard_scale,
        cmap_name=cmap_name,
        return_table=True
    )

    return list(set(res['interaction_group']))


import os
import numpy as np
from scipy.stats import rankdata
from scipy.special import binom #faster than comb
from ._df_apply import gen_row,gen_col,apply
from functools import partial



chunkSize = 100

#Function to index flat matrix as squareform matrix
def dist_index(i, j, matrix):
    
    if i == j:
        return(0.0)
    
    l = len(matrix)
    n = 0.5*(np.sqrt((8*l)+1)+1)
    index = int(l - binom(n-min(i, j), 2) + (max(i, j) - min(i, j) - 1))
    
    return(matrix[index])
    

#Function to index flat matrix as squareform matrix
def dist_multi_index(_array, matrix):
    
    results = np.zeros((len(_array), len(_array)))
    for i in range(len(_array)):
        for j in range(i, len(_array)):
            score = dist_index(_array[i], _array[j], matrix)
            results[i,j] = score
            results[j,i] = score
    
    return(results)
    

#Function to index rows of flat matrix as squareform matrix
def get_rows(_array, matrix):
    
    l = len(matrix)
    n = int(0.5*(np.sqrt((8*l)+1)+1))
    
    if _array.dtype != "bool":
        results = np.zeros((len(_array), n))
    
        for row, i in enumerate(_array):
            for j in range(n):
                index = int(l - binom(n - min(i, j), 2) + (max(i, j) - min(i, j) - 1))
                results[row,j] = matrix[index]
    
        return(results)
        
    else:
        results = np.zeros((np.sum(_array), n))
        row = 0
        for i, b in enumerate(_array):
            if b == True:
                for j in range(n):
                    index = int(l - binom(n - min(i, j), 2) + (max(i, j) - min(i, j) - 1))
                    results[row, j] = matrix[index]
                
                row += 1
                
        return(results)
    
    
#The following are supporting function for GetClusters.
def CoreSize(BranchSize, minClusterSize):
    
    BaseCoreSize = minClusterSize / 2 + 1
    if BaseCoreSize < BranchSize:
        CoreSize = BaseCoreSize + np.sqrt(BranchSize - BaseCoreSize)
    else:
        CoreSize = BranchSize
    
    return(int(CoreSize))
    

# This assumes the diagonal of the distance matrix
# is zero, BranchDist is a square matrix whose dimension is at least 2.
def CoreScatter(BranchDist, minClusterSize):
    
    nPoints = BranchDist.shape[0]
    PointAverageDistances = np.sum(BranchDist, axis=1) / (nPoints - 1)
    CoreSize = minClusterSize / 2 + 1

    if CoreSize < nPoints:
        EffCoreSize = CoreSize + np.sqrt(nPoints - CoreSize)
        order = np.argsort(PointAverageDistances)
        Core = order[np.arange(EffCoreSize)]
    else:
        Core = np.arange(nPoints)
        EffCoreSize = nPoints

    CoreAverageDistances = np.sum(BranchDist[Core, Core], axis=1) / (EffCoreSize - 1)
    
    return(np.mean(CoreAverageDistances))
    

def interpolate(data, index):
    
    i = np.round(index)
    n = len(data)
    if i < 0: return(data[0])
    if i >= n: return(data[-1])

    r = index - i
        
    return(data[i-1] * (1 - r) + data[i] * r)


def cutreeHybrid(link, distM,
                 cutHeight = None, minClusterSize = 20, deepSplit = 1,
                 maxCoreScatter = None, minGap = None,
                 maxAbsCoreScatter = None, minAbsGap = None,
                 minSplitHeight = None, minAbsSplitHeight = None,
                 externalBranchSplitFnc = None, minExternalSplit = None,
                 externalSplitOptions = [],
                 externalSplitFncNeedsDistance = None,
                 assumeSimpleExternalSpecification = True,
                 pamStage = True, pamRespectsDendro = True,
                 useMedoids = False,
                 maxPamDist = None,
                 respectSmallClusters = True,
                 verbose = 2, indent = 0):
    
    
    dendro_height = get_heights(link)
    dendro_merge = get_merges(link)

    if maxPamDist == None:
        maxPamDist = cutHeight
        
    nMerge = len(dendro_height)
    refQuantile = 0.05
    refMerge = np.round(nMerge * refQuantile)

    if refMerge < 1: refMerge = 1

    refHeight = dendro_height[int(refMerge) - 1]

    if cutHeight == None:
        cutHeight = 0.99 * (np.max(dendro_height) - refHeight) + refHeight
        print("..cutHeight not given, setting it to", cutHeight, 
              " ===>  99% of the (truncated) height range in dendro.")
    else:
        if cutHeight > np.max(dendro_height): cutHeight = np.max(dendro_height)
    
    if maxPamDist == None: maxPamDist = cutHeight

    nMergeBelowCut = np.sum(dendro_height <= cutHeight)

    if nMergeBelowCut < minClusterSize:
        print("cutHeight set too low; no merges below the cut.")
        return(np.zeros(nMerge+1))
    
    # fill in this section once understood better
    if externalBranchSplitFnc != None:
        nExternalSplits = len(externalBranchSplitFnc)
        if len(minExternalSplit) < 1:
            raise AttributeError("minExternalBranchSplit must be given.")
        if assumeSimpleExternalSpecification and nExternalSplits == 1:
            pass
    else:
        nExternalSplits = 0


    MxBranches = nMergeBelowCut
    branch_isBasic = np.repeat(True, MxBranches)
    branch_isTopBasic = np.repeat(True, MxBranches)
    branch_failSize = np.repeat(False, MxBranches)
    branch_rootHeight = np.repeat(np.nan, MxBranches)
    branch_size = np.repeat(2, MxBranches)
    branch_nMerge = np.repeat(1, MxBranches)
    branch_nSingletons = np.repeat(2, MxBranches)
    branch_nBasicClusters = np.repeat(0, MxBranches)
    branch_mergedInto = np.repeat(0, MxBranches)
    branch_attachHeight = np.repeat(np.nan, MxBranches)
    #branch_singletons = np.zeros(MxBranches)
    branch_singletons = [np.nan] * MxBranches
    #branch_basicClusters = pd.Series(np.zeros(MxBranches))
    branch_basicClusters = [np.nan] * MxBranches
    #branch_mergingHeights = pd.Series(np.zeros(MxBranches))
    branch_mergingHeights = [np.nan] * MxBranches
    #branch_singletonHeights = pd.Series(np.zeros(MxBranches))
    branch_singletonHeights = [np.nan] * MxBranches


    nBranches = 0

    spyIndex = None
    if os.path.isfile(".dynamicTreeCutSpyFile"):
        spyIndex = pd.read_csv(".dynamicTreeCutSpyFile")
        print("Found 'spy file' with indices of objects to watch for.")
        spyIndex = spyIndex.iloc[:,1].values
    

    defMCS = np.array([0.64, 0.73, 0.82, 0.91, 0.95])
    defMG = (1 - defMCS) * 3 / 4.0

    nSplitDefaults = len(defMCS)

    if type(deepSplit) == bool: deepSplit = int(deepSplit) * (nSplitDefaults - 2)
    deepSplit = deepSplit + 1

    if deepSplit < 1 or deepSplit > nSplitDefaults:
        raise IndexError("Parameter deepSplit (value", deepSplit,
                         ") out of range: allowable range is 0 through",
                         nSplitDefaults - 1)
    
    if maxCoreScatter == None: maxCoreScatter = interpolate(defMCS, deepSplit)
    if minGap == None: minGap = interpolate(defMG, deepSplit)

    if maxAbsCoreScatter == None:
        maxAbsCoreScatter = refHeight + maxCoreScatter * (cutHeight - refHeight)
    if minAbsGap == None:
        minAbsGap = minGap * (cutHeight - refHeight)

    if minSplitHeight == None: minSplitHeight = 0

    if minAbsSplitHeight == None:
        minAbsSplitHeight = refHeight + minSplitHeight * (cutHeight - refHeight)
    
    nPoints = nMerge + 1

    IndMergeToBranch = np.repeat(0, nMerge)

    onBranch = np.repeat(0, nPoints)

    RootBranch = 0

    mergeDiagnostics = dict(smI = np.repeat(np.nan, nMerge), smSize = np.repeat(np.nan, nMerge), 
                            smCrSc = np.repeat(np.nan, nMerge), smGap = np.repeat(np.nan, nMerge), 
                            lgI = np.repeat(np.nan, nMerge), lgSize = np.repeat(np.nan, nMerge), 
                            lgCrSc = np.repeat(np.nan, nMerge), lgGap = np.repeat(np.nan, nMerge),
                            merged = np.repeat(np.nan, nMerge))

    if nExternalSplits > 0:
        #externalMergeDiags = pd.DataFrame(np.repeat(np.nan, nMerge*nExternalSplits).reshape(nMerge, nExternalSplits))
        #externalMergeDiags.columns = paste("externalBranchSplit", nExternalSplits, sep = ".")
        pass

    extender = np.zeros(chunkSize, dtype=int)
    
    for merge in range(nMerge):
        if dendro_height[merge] <= cutHeight:
            # are both merged objects singletons?
            if dendro_merge[merge, 0] < 0 and dendro_merge[merge, 1] < 0:
                nBranches = nBranches + 1
                branch_isBasic[nBranches - 1] = True
                branch_isTopBasic[nBranches - 1] = True
                branch_singletons[nBranches - 1] = np.append(-dendro_merge[merge,], extender)
                branch_basicClusters[nBranches - 1] = extender
                branch_mergingHeights[nBranches - 1] = np.append(np.repeat(dendro_height[merge], 2), extender)
                branch_singletonHeights[nBranches - 1] = np.append(np.repeat(dendro_height[merge], 2), extender)
                IndMergeToBranch[merge] = nBranches
                RootBranch = nBranches
            elif sign(dendro_merge[merge,0]) * sign(dendro_merge[merge,1]) < 0:
                clust = IndMergeToBranch[int(np.max(dendro_merge[merge,])) - 1]
                if clust == 0: raise ValueError("a previous merge has no associated cluster. Sorry!")
                gene = -np.min(dendro_merge[merge,])
                ns = branch_nSingletons[clust - 1] + 1
                nm = branch_nMerge[clust - 1] + 1
                if branch_isBasic[clust - 1]:
                    if ns > len(branch_singletons[clust - 1]):
                        branch_singletons[clust - 1] = np.append(branch_singletons[clust - 1], extender)
                        branch_singletonHeights[clust - 1] = np.append(branch_singletonHeights[clust - 1], extender)
                    branch_singletons[clust - 1][ns - 1] = gene
                    branch_singletonHeights[clust - 1][ns - 1] = dendro_height[merge]
                else:
                    onBranch[int(gene) - 1] = clust
                
                if nm >= len(branch_mergingHeights[clust - 1]):
                    branch_mergingHeights[clust - 1] = np.append(branch_mergingHeights[clust - 1], extender)
                branch_mergingHeights[clust - 1][nm - 1] = dendro_height[merge]
                branch_size[clust - 1] = branch_size[clust - 1] + 1
                branch_nMerge[clust - 1] = nm
                branch_nSingletons[clust - 1] = ns
                IndMergeToBranch[merge] = clust
                RootBranch = clust
            else:
                # attempt to merge two branches:
                clusts = IndMergeToBranch[dendro_merge[merge,] - 1]
                sizes = branch_size[clusts - 1]
                # Note: for 2 elements, rank and order are the same.
                rnk = rankdata(sizes, method = "ordinal")
                small = clusts[rnk[0] - 1]
                large = clusts[rnk[1] - 1]
                sizes = sizes[rnk - 1]
                branch1 = np.nan if np.any(np.isnan(branch_singletons[large - 1])) else branch_singletons[large - 1][np.arange(sizes[1])]
                branch2 = np.nan if np.any(np.isnan(branch_singletons[small - 1])) else branch_singletons[small - 1][np.arange(sizes[0])]
                spyMatch = False
                if spyIndex != None:
                    n1 = len(set(branch1) & set(spyIndex))
                    if n1 / len(branch1) > 0.99 and n1 / len(spyIndex) > 0.99:
                        print("Found spy match for branch 1 on merge", merge)
                        spyMatch = True
                    n2 = len(set(branch2) & set(spyIndex))
                    if n2 / len(branch1) > 0.99 and n2 / len(spyIndex) > 0.99:
                        print("Found spy match for branch 2 on merge", merge)
                        spyMatch = True
            
                if branch_isBasic[small - 1]:
                    coresize = CoreSize(branch_nSingletons[small - 1], minClusterSize)
                    Core = np.array(branch_singletons[small - 1][np.arange(int(coresize))], dtype=int)
                    # SmAveDist = mean(apply(distM[Core, Core], 2, sum)/(coresize-1))
                    SmAveDist = np.mean(np.sum(dist_multi_index(Core - 1, distM), axis=1) / (coresize - 1))     
                else:
                    SmAveDist = 0
            
                if branch_isBasic[large - 1]:
                    coresize = CoreSize(branch_nSingletons[large - 1], minClusterSize)
                    Core = np.array(branch_singletons[large - 1][np.arange(int(coresize))], dtype=int)
                    LgAveDist = np.mean(np.sum(dist_multi_index(Core - 1, distM), axis=1) / (coresize -1 ))
                else:
                    LgAveDist = 0
                
                for key in mergeDiagnostics:
                    if key == "smI":
                        mergeDiagnostics[key][merge] = small
                    elif key == "smSize":
                        mergeDiagnostics[key][merge] = branch_size[small - 1]
                    elif key == "smCrSc":
                        mergeDiagnostics[key][merge] = SmAveDist
                    elif key == "smGap":
                        mergeDiagnostics[key][merge] = dendro_height[merge] - SmAveDist
                    elif key == "lgI":
                        mergeDiagnostics[key][merge] = large
                    elif key == "lgSize":
                        mergeDiagnostics[key][merge] = branch_size[large - 1]
                    elif key == "lgCrSc":
                        mergeDiagnostics[key][merge] = LgAveDist
                    elif key == "lgGap":
                        mergeDiagnostics[key][merge] = dendro_height[merge] - LgAveDist
                    elif key == "merged":
                        mergeDiagnostics[key][merge] = np.nan
                        
            
                # We first check each cluster separately for being too small, too diffuse, or too shallow:
                SmallerScores = [branch_isBasic[small - 1], 
                                 branch_size[small - 1] < minClusterSize,
                                 SmAveDist > maxAbsCoreScatter, 
                                 dendro_height[merge] - SmAveDist < minAbsGap,
                                 dendro_height[merge] < minAbsSplitHeight]
            
                if SmallerScores[0] * np.sum(SmallerScores[1:]) > 0:
                    DoMerge = True
                    SmallerFailSize = ~np.logical_or(SmallerScores[2], SmallerScores[3])  # Smaller fails only due to size
                else:
                    LargerScores = [branch_isBasic[large - 1], 
                                    branch_size[large - 1] < minClusterSize,
                                    LgAveDist > maxAbsCoreScatter, 
                                    dendro_height[merge] - LgAveDist < minAbsGap,
                                    dendro_height[merge] < minAbsSplitHeight]
                    if LargerScores[0] * np.sum(LargerScores[1:]) > 0:
                        # Actually: the large one is the one to be merged
                        DoMerge = True
                        SmallerFailSize = ~np.logical_or(LargerScores[2], LargerScores[3])  # cluster fails only due to size
                        x = small
                        small = large
                        large = x
                        sizes = sizes[::-1]
                    else:
                        DoMerge = False # None of the two satisfies merging criteria
            
                if DoMerge:
                    mergeDiagnostics["merged"][merge] = 1

                if ~DoMerge and nExternalSplits > 0 and branch_isBasic[small - 1] and branch_isBasic[large - 1]:
                    if verbose > 4: print("Entering external split code on merge ", merge)
                    branch1 = branch_singletons[large - 1][np.arange(sizes[1])]
                    branch2 = branch_singletons[small - 1][np.arange(sizes[0])]

                    if verbose > 4 or spyMatch: print("  ..branch lengths: ", sizes[0], ", ", sizes[1])
                    #if (any(is.na(branch1)) || any(branch1==0)) browser();
                    #if (any(is.na(branch2)) || any(branch2==0)) browser();
                
                
                    ##### fix after External Splits is understood better
                    es = 0
                    while es < nExternalSplits and ~DoMerge:
                        es = es + 1
                        args = externalSplitOptions[es - 1]
                        args = [args, list(branch1 = branch1, branch2 = branch2)]
                        #extSplit = do.call(externalBranchSplitFnc[es], args)
                        if spyMatch:
                            print(" .. external criterion ", es, ": ", extSplit)
                        DoMerge = extSplit < minExternalSplit[es - 1]
                        externalMergeDiags[merge, es - 1] = extSplit
                        if DoMerge:
                            mergeDiagnostics_merged[merge] = 2
                        else:
                            mergeDiagnostics_merged[merge] = 0
            
                if DoMerge:
                    # merge the small into the large cluster and close it.
                    branch_failSize[small - 1] = SmallerFailSize
                    branch_mergedInto[small - 1] = large
                    branch_attachHeight[small - 1] = dendro_height[merge]
                    branch_isTopBasic[small - 1] = False
                    nss = branch_nSingletons[small - 1]
                    nsl = branch_nSingletons[large - 1]
                    ns = nss + nsl
                
                    if branch_isBasic[large - 1]: 
                        nExt = np.ceil(  (ns - len(branch_singletons[large - 1])) / chunkSize  )
                    
                        if nExt > 0:
                            if verbose > 5:
                                print("Extending singletons for branch", large, "by", nExt, " extenders.")
                        
                            branch_singletons[large - 1] = np.append(branch_singletons[large - 1], np.repeat(extender, nExt))
                            branch_singletonHeights[large - 1] = np.append(branch_singletonHeights[large - 1], np.repeat(extender, nExt))
                    
                        branch_singletons[large - 1][np.arange(nsl,ns)] = branch_singletons[small - 1][np.arange(nss)]
                        branch_singletonHeights[large - 1][np.arange(nsl,ns)] = branch_singletonHeights[small - 1][np.arange(nss)]
                        branch_nSingletons[large - 1] = ns
                    
                    else:
                        if ~branch_isBasic[small - 1]:
                            raise ValueError("merging two composite clusters. Sorry!")
                    
                        onBranch[ branch_singletons[small - 1][branch_singletons[small - 1] != 0] - 1 ] = large
                
                    nm = branch_nMerge[large - 1] + 1
                
                    if nm > len(branch_mergingHeights[large - 1]):
                        branch_mergingHeights[large - 1] = np.append(branch_mergingHeights[large - 1], extender)
                
                    branch_mergingHeights[large - 1][nm - 1] = dendro_height[merge]
                    branch_nMerge[large - 1] = nm
                    branch_size[large - 1] = branch_size[small - 1] + branch_size[large - 1]
                    IndMergeToBranch[merge] = large
                    RootBranch = large
                else:
                    # start or continue a composite cluster.

                    # If large is basic and small is not basic, switch them.
                    if branch_isBasic[large - 1] and ~branch_isBasic[small - 1]:
                        x = large
                        large = small
                        small = x
                        sizes = sizes[::-1]
                
                    # Note: if pamRespectsDendro, need to start a new composite cluster every time two branches merge,
                    # otherwise will not have the necessary information.
                    # Otherwise, if the large cluster is already composite, I can simply merge both clusters into 
                    # one of the non-composite clusters.

                    if branch_isBasic[large - 1] or pamStage and pamRespectsDendro:
                        nBranches = nBranches + 1
                        branch_attachHeight[[large - 1, small - 1]] = dendro_height[merge]
                        branch_mergedInto[[large - 1, small - 1]] = nBranches
                        if branch_isBasic[small - 1]:
                            addBasicClusters = small # add basic clusters
                        else:
                            addBasicClusters = branch_basicClusters[small - 1]
                        if branch_isBasic[large - 1]:
                            addBasicClusters = np.append(addBasicClusters, large)
                        else:
                            addBasicClusters = np.append(addBasicClusters, branch_basicClusters[large - 1])
                        # print(paste("  Starting a composite cluster with number", nBranches));
                        branch_isBasic[nBranches - 1] = False
                        branch_isTopBasic[nBranches - 1] = False
                        branch_basicClusters[nBranches - 1] = addBasicClusters
                        branch_mergingHeights[nBranches - 1] = np.append(np.repeat(dendro_height[merge], 2), extender)
                        branch_nMerge[nBranches - 1] = 2
                        branch_size[nBranches - 1] = np.sum(sizes)
                        branch_nBasicClusters[nBranches - 1] = len(addBasicClusters)
                        IndMergeToBranch[merge] = nBranches
                        RootBranch = nBranches
                    else:
                        # Add small branch to the large one 
                        addBasicClusters = small if branch_isBasic[small - 1] else branch_basicClusters[small - 1]
                        nbl = branch_nBasicClusters[large - 1]
                        #small might be an int
                        try:
                            nb = branch_nBasicClusters[large - 1] + len(addBasicClusters)
                        except TypeError:
                            nb = branch_nBasicClusters[large - 1] + 1
                                                
                        if nb > len(branch_basicClusters[large - 1]):
                            nExt = np.ceil(  ( nb - len(branch_basicClusters[large - 1])) / chunkSize)
                            branch_basicClusters[large - 1] = np.append(branch_basicClusters[large - 1], np.repeat(extender, nExt))
                    
                        branch_basicClusters[large - 1][np.arange(nbl,nb)] = addBasicClusters
                        branch_nBasicClusters[large - 1] = nb
                        branch_size[large - 1] = branch_size[large - 1] + branch_size[small - 1]
                        nm = branch_nMerge[large - 1] + 1
                    
                        if nm > len(branch_mergingHeights[large - 1]):
                            branch_mergingHeights[large - 1] = np.append(branch_mergingHeights[large - 1], extender) 
                    
                        branch_mergingHeights[large - 1][nm - 1] = dendro_height[merge]
                        branch_nMerge[large - 1] = nm
                        branch_attachHeight[small - 1] = dendro_height[merge]
                        branch_mergedInto[small - 1] = large
                        IndMergeToBranch[merge] = large
                        RootBranch = large
        
    if verbose > 2: print("..Going through detected branches and marking clusters..")
            
    isCluster = np.repeat(False, nBranches)
    SmallLabels = np.repeat(0, nPoints)

    for clust in range(nBranches):
    
        if np.isnan(branch_attachHeight[clust]): branch_attachHeight[clust] = cutHeight
        if branch_isTopBasic[clust]:
            coresize = CoreSize(branch_nSingletons[clust], minClusterSize)
            Core = branch_singletons[clust][np.arange(coresize)]
            CoreScatter = np.mean(np.sum(dist_multi_index(Core - 1, distM), axis=1) / (coresize - 1))
            isCluster[clust] = np.logical_and(np.logical_and(branch_isTopBasic[clust],
                                                             branch_size[clust] >= minClusterSize),
                                              np.logical_and(CoreScatter < maxAbsCoreScatter,
                                                             branch_attachHeight[clust] - CoreScatter > minAbsGap))
        else:
            CoreScatter = 0
        
        if branch_failSize[clust]: SmallLabels[branch_singletons[clust][branch_singletons[clust] != 0] - 1] = clust + 1
    
    if not respectSmallClusters: SmallLabels = np.repeat(0, nPoints)

    if verbose > 2: print(spaces, "..Assigning Tree Cut stage labels..")

    Colors = np.repeat(0, nPoints)
    coreLabels = np.repeat(0, nPoints)
    clusterBranches = np.arange(nBranches)[isCluster]
    branchLabels = np.repeat(0, nBranches)
    color = 0
                    
    for clust in clusterBranches:
        color = color + 1
        Colors[branch_singletons[clust][branch_singletons[clust] != 0] - 1] = color
        SmallLabels[branch_singletons[clust][branch_singletons[clust] != 0] - 1] = 0
        coresize = CoreSize(branch_nSingletons[clust], minClusterSize)
        Core = branch_singletons[clust][np.arange(coresize)]
        coreLabels[Core - 1] = color
        branchLabels[clust] = color

    Labeled = np.arange(nPoints)[Colors != 0]
    Unlabeled = np.arange(nPoints)[Colors == 0]
    nUnlabeled = len(Unlabeled)
    UnlabeledExist = nUnlabeled > 0

    if len(Labeled) > 0:
        LabelFac = factor(Colors[Labeled])
        nProperLabels = nlevels(LabelFac)
    else:
        nProperLabels = 0


    if pamStage and UnlabeledExist and nProperLabels > 0:
        if verbose > 2: print(spaces, "..Assigning PAM stage labels..")
        nPAMed = 0
        # Assign some of the grey genes to the nearest module. Define nearest as the distance to the medoid,
        # that is the point in the cluster that has the lowest average distance to all other points in the
        # cluster. First get the medoids.
        if useMedoids:
            Medoids = np.repeat(0, nProperLabels)
            ClusterRadii = np.repeat(0.0, nProperLabels)
            for cluster in range(1, nProperLabels + 1):
                InCluster = np.arange(1,nPoints+1)[Colors == cluster]
                DistInCluster = dist_multi_index(InCluster - 1, distM)
                #DistInCluster = distM[InCluster, InCluster]
                DistSums = np.sum(DistInCluster, axis=1)
                Medoids[cluster - 1] = InCluster[np.argmin(DistSums)]
                ClusterRadii[cluster - 1] = np.max(DistInCluster[:, np.argmin(DistSums)])
            # If small clusters are to be respected, assign those first based on medoid-medoid distances.
            if respectSmallClusters:
                FSmallLabels = factor(SmallLabels)
                SmallLabLevs = levels(FSmallLabels)
                nSmallClusters = nlevels(FSmallLabels) - (SmallLabLevs[0] == 0)
                if nSmallClusters > 0 :
                    for sclust in SmallLabLevs[SmallLabLevs != 0]:
                        InCluster = np.arange(nPoints)[SmallLabels == sclust]
                        if pamRespectsDendro:
                            onBr = np.unique(onBranch[InCluster])
                            if len(onBr) > 1:
                                raise ValueError("Internal error: objects in a small cluster are marked to belong",
                                                 "\nto several large branches:")
                            if onBr > 0:
                                basicOnBranch = branch_basicClusters[onBr[0] - 1]
                                labelsOnBranch = branchLabels[basicOnBranch - 1]
                            else:
                                labelsOnBranch = None
                        else:
                            labelsOnBranch = np.arange(1, nProperLabels + 1)
                        # printFlush(paste("SmallCluster", sclust, "has", length(InCluster), "elements."));
                        DistInCluster = dist_multi_index(InCluster, distM)
                        #DistInCluster = distM[InCluster, InCluster]
                        if len(labelsOnBranch) > 0:
                            if len(InCluster) > 1:
                                DistSums = apply(np.sum, DistInCluster, 1)
                                smed = InCluster[np.argmin(DistSums)]
                                DistToMeds = get_rows(Medoids[labelsOnBranch - 1][Medoids[labelsOnBranch - 1] != 0] - 1, distM)[:, smed]
                                closest = np.argmin(DistToMeds)
                                DistToClosest = DistToMeds[closest]
                                closestLabel = labelsOnBranch[closest]
                                if DistToClosest < ClusterRadii[closestLabel - 1] or DistToClosest <  maxPamDist:
                                    Colors[InCluster] = closestLabel
                                    nPAMed = nPAMed + len(InCluster)
                                else: Colors[InCluster] = -1  # This prevents individual points from being assigned later 
                        else:
                            Colors[InCluster] = -1
         
                # Assign leftover unlabeled objects to clusters with nearest medoids
                Unlabeled = np.arange(nPoints)[Colors == 0]
                if len(Unlabeled > 0):
                    for obj in Unlabeled:
                        if pamRespectsDendro:
                            onBr = onBranch[obj]
                            if onBr > 0:
                                basicOnBranch = branch_basicClusters[onBr - 1]
                                labelsOnBranch = branchLabels[basicOnBranch - 1]
                            else:
                                labelsOnBranch = None
                        else:
                            labelsOnBranch = np.arange(nProperLabels)
                        if labelsOnBranch != None:
                            UnassdToMedoidDist = get_rows(Medoids[labelsOnBranch - 1] - 1, distM)[:,obj]
                            #UnassdToMedoidDist = distM[Medoids[labelsOnBranch], obj]
                            nearest= np.argmin(UnassdToMedoidDist)
                            NearestCenterDist = UnassdToMedoidDist[nearest]
                            nearestMed = labelsOnBranch[nearest]
                            if NearestCenterDist < ClusterRadii[nearestMed - 1] or NearestCenterDist < maxPamDist:
                                Colors[obj] = nearestMed
                                nPAMed = nPAMed + 1

                UnlabeledExist = np.sum(Colors == 0) > 0
        else: # Instead of medoids, use average distances
        # This is the default method, so I will try to tune it for speed a bit.
            ClusterDiam = np.repeat(0, nProperLabels)
            for cluster in range(nProperLabels):
                InCluster = np.arange(nPoints)[Colors == cluster]
                nInCluster = len(InCluster)
                DistInCluster = dist_multi_index(InCluster, distM)
                #DistInCluster = distM[InCluster, InCluster]
                if nInCluster > 1:
                    AveDistInClust = np.sum(DistInCluster, axis=1) / (nInCluster - 1)
                    ClusterDiam[cluster] = np.max(AveDistInClust)
                else:
                    ClusterDiam[cluster] = 0

            # If small clusters are respected, assign them first based on average cluster-cluster distances.
            ColorsX = Colors.copy()
            if respectSmallClusters:
                FSmallLabels = factor(SmallLabels) #### think about
                SmallLabLevs = levels(FSmallLabels) ##### think about
                nSmallClusters = nlevels(FSmallLabels) - (SmallLabLevs[0] == 0)
                if nSmallClusters > 0:
                    if pamRespectsDendro:
                        for sclust in SmallLabLevs[SmallLabLevs != 0]:
                            InCluster = np.arange(nPoints)[SmallLabels == sclust]
                            onBr = np.unique(onBranch[InCluster])

                            if len(onBr) > 1:
                                raise ValueError("objects in a small cluster are marked to belong",
                                                 "\nto several large branches:")
                            if onBr > 0:
                                basicOnBranch = branch_basicClusters[onBr[0] - 1]
                                labelsOnBranch = branchLabels[basicOnBranch - 1]
                                useObjects = np.in1d(ColorsX, np.unique(labelsOnBranch))
                                DistSClustClust = get_rows(InCluster, distM)[:,useObjects]
                                #DistSClustClust = distM[InCluster, useObjects]
                                MeanDist = np.mean(DistSClustClust, axis=0)
                                useColorsFac = factor(ColorsX[useObjects]) ### think about
                                MeanMeanDist = tapply(MeanDist, useColorsFac, np.mean) ## think about
                                nearest = np.argmin(MeanMeanDist)
                                NearestDist = MeanMeanDist[nearest]
                                nearestLabel = levels(useColorsFac)[nearest] ## think about
                                if NearestDist < ClusterDiam[nearestLabel - 1] or NearestDist <  maxPamDist:
                                    Colors[InCluster] = nearestLabel
                                    nPAMed = nPAMed + len(InCluster)
                                else:
                                    Colors[InCluster] = -1  # This prevents individual points from being assigned later
 
                    else:
                        labelsOnBranch = np.arange(nProperLabels)
                        useObjects = np.arange(nPoints)[ColorsX != 0]
                        for sclust in SmallLabLevs[SmallLabLevs != 0]:
                            InCluster = np.arange(nPoints)[SmallLabels == sclust]
                            DistSClustClust = get_rows(InCluster, distM)[:,useObjects]
                            #DistSClustClust = distM[InCluster, useObjects]
                            MeanDist = np.mean(DistSClustClust, axis=0)
                            useColorsFac = factor(ColorsX[useObjects]) ### think about
                            MeanMeanDist = tapply(MeanDist, useColorsFac, np.mean) ### think about
                            nearest = np.argmin(MeanMeanDist)
                            NearestDist = MeanMeanDist[nearest]
                            nearestLabel = levels(useColorsFac)[nearest] ## think about
                            if NearestDist < ClusterDiam[nearestLabel - 1] or NearestDist <  maxPamDist:
                                Colors[InCluster] = nearestLabel
                                nPAMed = nPAMed + len(InCluster)
                            else:
                                Colors[InCluster] = -1  # This prevents individual points from being assigned later

            # Assign leftover unlabeled objects to clusters with nearest medoids
            Unlabeled = np.arange(nPoints)[Colors == 0]
            #ColorsX = Colors;
            if len(Unlabeled) > 0:
                if pamRespectsDendro:
                    unlabOnBranch = Unlabeled[onBranch[Unlabeled] > 0]
                    for obj in unlabOnBranch:
                        onBr = onBranch[obj]
                        basicOnBranch = branch_basicClusters[onBr - 1]
                        labelsOnBranch = branchLabels[basicOnBranch - 1]
                        useObjects = np.in1d(ColorsX, np.unique(labelsOnBranch))
                        useColorsFac = factor(ColorsX[useObjects]) ### think about
                        #UnassdToClustDist = tapply(distM[useObjects, obj], useColorsFac, mean) ### think about
                        UnassdToClustDist = tapply(get_rows(useObjects, distM)[:,obj], useColorsFac, np.mean) ### think about
                        nearest = np.argmin(UnassdToClustDist)
                        NearestClusterDist = UnassdToClustDist[nearest]
                        nearestLabel = levels(useColorsFac)[nearest] ### think about
                        if NearestClusterDist < ClusterDiam[nearestLabel - 1] or NearestClusterDist < maxPamDist:
                            Colors[obj] = nearestLabel
                            nPAMed = nPAMed + 1

                else:
                    useObjects = np.arange(nPoints)[ColorsX != 0]
                    useColorsFac = factor(ColorsX[useObjects]) ## think about
                    nUseColors = nlevels(useColorsFac) ### think about
                    UnassdToClustDist = tapply_df(get_rows(useObjects, distM)[:,Unlabeled], useColorsFac, np.mean, 1)
                    #UnassdToClustDist = df_apply.apply(distM[useObjects, Unlabeled], 1, tapply, useColorsFac, mean) ### think about
                    # Fix dimensions for the case when there's only one cluster
                    #dim(UnassdToClustDist) = np.append(nUseColors, len(Unlabeled)) ### think about
                    nearest = apply(np.argmin, UnassdToClustDist, 1)
                    nearestDist = apply(np.min, UnassdToClustDist, 1)
                    nearestLabel = levels(useColorsFac)[nearest - 1] ### think about
                    assign = np.logical_or(nearestDist < ClusterDiam[nearestLabel - 1], nearestDist < maxPamDist)
                    Colors[Unlabeled[assign]] = nearestLabel[assign]
                    nPAMed = nPAMed + np.sum(assign)

        if verbose > 2: print("....assigned", nPAMed, "objects to existing clusters.")

    
    # Relabel labels such that 1 corresponds to the largest cluster etc.
    Colors[Colors < 0] = 0
    UnlabeledExist = np.sum(Colors == 0) > 0
    NumLabs = Colors + 1
    Sizes = table(NumLabs) ### think about
    Sizes.index=np.arange(len(Sizes))

    if UnlabeledExist:
        if len(Sizes) > 1:
            SizeRank = np.append(1, rankdata(-Sizes[1:len(Sizes)], method="ordinal")+1)
        else:
            SizeRank = 1
        OrdNumLabs = SizeRank[NumLabs - 1]
    else:
        SizeRank = rankdata(-Sizes[np.arange(len(Sizes))], method="ordinal")
        OrdNumLabs = SizeRank[NumLabs - 2]
    ordCoreLabels = OrdNumLabs - UnlabeledExist
    ordCoreLabels[coreLabels == 0] = 0

    if verbose > 0: print( "..done.")

    results = dict(labels = OrdNumLabs-UnlabeledExist,
                   cores = ordCoreLabels,
                   smallLabels = SmallLabels,
                   onBranch = onBranch,
                   mergeDiagnostics = mergeDiagnostics if nExternalSplits==0 else pd.DataFrame({'x':mergeDiagnostics, 'y':externalMergeDiags}),
                   mergeCriteria = dict(maxCoreScatter = maxCoreScatter, minGap = minGap, 
                                        maxAbsCoreScatter = maxAbsCoreScatter, minAbsGap = minAbsGap, 
                                        minExternalSplit = minExternalSplit),
                   branches  = dict(nBranches = nBranches, # Branches = Branches, 
                                    IndMergeToBranch = IndMergeToBranch,
                                    RootBranch = RootBranch, isCluster = isCluster, 
                                    nPoints = nMerge+1))

    return(results)












import pandas as pd
import numpy as np
from scipy.cluster.hierarchy import to_tree


def sign(value):
    #python version of R's sign
    
    if value > 0:
        return(1)
    elif value < 0:
        return(-1)
    else:
        return(0)

def paste(string, n, sep=""):
    #python version of R's paste
    
    results = []
    for i in range(n):
        results.append(string + sep + str(i))
    
    return(results)


def get_heights(Z):
    #python verison of R's dendro$height
    #height = np.zeros(len(dendro["dcoord"]))
    
    #for i, d in enumerate(dendro["dcoord"]):
        #height[i] = d[1]
    
    clusternode = to_tree(Z, True)
    #height = np.array([c.dist for c in clusternode[1]])
    height = np.array([c.dist for c in clusternode[1] if c.is_leaf() != True])
        
    #height.sort()
    
    return(height)
    

def get_merges(z):
    #python version of R's dendro$merge
    n = z.shape[0]
    merges = np.zeros((z.shape[0], 2), dtype=int)
        
    for i in range(z.shape[0]):
        for j in range(2):
            if z[i][j] <= n:
                merges[i][j] = -(z[i][j] + 1)
            else:
                cluster = z[i][j] - n
                merges[i][j] = cluster
                
    return(merges)
    
    
def factor(vector):
    return(vector)


def nlevels(vector):
    #python version of R's nlevels
    return(len(np.unique(vector)))


def levels(vector):
    #python version of R's levels
    return(np.unique(vector))


def tapply(vector, index, function): #can add **args, **kwargs
    #python version of R's tapply
    
    factors = np.unique(index)
    
    #results = pd.Series(np.repeat(np.nan, len(factors)))
    results = np.repeat(np.nan, len(factors))
    #results.index = factors
    
    for i, k in enumerate(factors):
        subset = vector[index == k]
        #results.iloc[i] = function(subset)
        results[i] = function(subset)
    
    return(results)


def tapply_df(df, index, function, axis=0): #can add **args, **kwargs
    #python version of R's tapply
    
    factors = np.unique(index)
    
    if axis == 1:
        #results = pd.DataFrame(np.zeros((len(factors), df.shape[1])))
        results = np.zeros((len(factors), df.shape[1]))
    else:
        #results = pd.DataFrame(np.zeros((df.shape[0], len(factors))))
        results = np.zeros((df.shape[0], len(factors)))
    
    #results.index = factors
    
    if axis == 1:
        for j in range(df.shape[1]):
            for i, k in enumerate(factors):
                subset = df[index == k, j]
                #results.iloc[i, j] = function(subset)
                results[i, j] = function(subset)
    else:
        for i in range(df.shape[0]):
            for j, k in enumerate(factors):
                subset = df[i, index == k]
                #results.iloc[i, j] = function(subset)
                results[i, j] = function(subset)
    
    return(results)


def table(vector):
    
    factors = np.unique(vector)
    results = pd.Series(np.zeros(len(factors), dtype=int))
    results.index = factors
    
    for i, k in enumerate(factors):
        results.iloc[i] = np.sum(vector == k)
        
    return(results)











r"""
Pyomic (A omic framework for multi-omic analysis)
"""

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from pkg_resources import get_distribution
    version = lambda name: get_distribution(name).version

from . import bulk,single,mofapy2,utils,bulk2single,pp

name = "omicverse"
__version__ = version(name)





from __future__ import division
import sys
import scipy as s
import numpy as np
import pandas as pd
# from fastdtw import fastdtw
import copy
# import gpytorch
# import torch
# from mofapy2.core.distributions.multi_task_GP import MultitaskGPModel, ELBO, myMultitaskGaussianLikelihood
# from gpytorch.likelihoods import MultitaskGaussianLikelihood
from .. import gp_utils
from .variational_nodes import *
from .Kc_node import Kc_Node
from .Kg_node import Kg_Node
from ..gp_utils import *
from .. import gpu_utils


# TODO:
# - implement warping for more than one covariate
# - test sparse + warping

class Sigma_Node_base(Node):
    """
    Sigma node to optimises the GP hyperparameters for each factor and
    perform alignment of covariates per group.

    The covariance matrix is modelled as a Kronecker product of a (low-rank) group kernel
    and a covariate kernel where possible:
    Sigma = (1-zeta) * KG \otimes KC + zeta * I

    Or as an elementwise multiplication of the two kernels, where no complete sampling is available.

    PARAMETERS
    ----------
    dim: dimensionality of the node (= number of latent factors)
    sample_cov: covariates for construction of the covariance matrix from distances (array of length G x C)
    groups: group label of each observation (array of length G)
    start_opt: in which iteration to start with optimizing the GP hyperparameters
    opt_freq: how often to hyperparameter optimization, every n-th iteration
    n_grid: number of grid points to optimize the lengthscale on
    rankx: rank of group covariance matrix \sum_rank x^T x, default: 1
    model_groups: whether to use a group kernel on top of the covariate kernel? If False, the group kernel corresponds to 1 for each pair of group.
    """

    def __init__(self, dim, sample_cov, groups, start_opt=20, opt_freq = 10,
                 n_grid = 10, rankx = None,  model_groups = False):
        super().__init__(dim)

        # dimensions and inputs
        self.use_gradients = model_groups                           # without group model there is only a single parameter (zeta)
        self.mini_batch = None
        self.sample_cov = sample_cov
        self.group_labels = groups
        self.N = sample_cov.shape[0]                                # total number of observation
        self.K = dim[0]                                             # number of factors

        # compatibility to other Sigma nodes and avoid code duplication
        self.sample_cov_transformed = copy.copy(sample_cov)         # for compatibility with warping node

        # hyperparameter optimization
        self.start_opt = start_opt
        self.n_grid = n_grid
        self.iter = 0                                               # counter of iteration to keep track when to optimize lengthscales
        self.zeta = np.ones(self.K)                                 # noise hyperparameter, corresponds to idenity matrices (and the init of SimgaInv terms)
        self.gridix = np.zeros(self.K, dtype = np.int8)             # index of the lengthscale grid values to use per factor
        self.struct_sig = np.zeros(self.K)                          # store ELBO improvements compared to diagonal covariance
        self.opt_freq = opt_freq
        assert self.start_opt % self.opt_freq == 0,\
            "start_opt should be a multiple of opt_freq"            # to ensure in the first opt. step optimization is performed

        # initialize group kernel
        self.model_groups = model_groups
        self.groupsidx = pd.factorize(self.group_labels)[0]         # for each sample gives the idx in groups
        self.groups = np.unique(self.group_labels)                  # distinct group labels

    def initKc(self, transformed_sample_cov, cov4grid=None, spectral_decomp=True):
        """
        Method to initialize the components required for the covariate kernel
        """
        if self.model_groups:
            self.covariates = np.unique(transformed_sample_cov, axis=0)  # distinct covariate values
            # non unique if no group model
            self.covidx = np.asarray([np.where((self.covariates == transformed_sample_cov[j, :]).all(axis=1))[0].item()
                                      for j in range(self.Nu)])  # for each sample gives the idx in covariates
        else:
            self.covariates = transformed_sample_cov  # all covariate values

        self.C = self.covariates.shape[0]  # number of covariate values

        # set covariate kernel
        self.Kc = Kc_Node(dim=(self.K, self.C), covariates=self.covariates, n_grid=self.n_grid, cov4grid=cov4grid,
                          spectral_decomp=spectral_decomp)

    def initKg(self, rank, spectral_decomp):
        """
        Method to initialize the group kernel
        """
        # set group kernel
        self.Kg = Kg_Node(dim=(self.K, self.G), rank=rank, spectral_decomp=spectral_decomp)

    def precompute(self, options):
        gpu_utils.gpu_mode = options['gpu_mode']

    def get_components(self, k):
        """
        Method to fetch ELBO-optimal covariance matrix components for a given factor k
        """
        Vc, Dc = self.Kc.get_kernel_components_k(k)

        if self.model_groups:
            Vg, Dg = self.Kg.get_kernel_components_k(k)
        else:
            Vg = np.array([1]);
            Dg = np.array([1])

        zeta = self.get_zeta()[k]

        return {'Vc': Vc, 'Vg': Vg, 'Dc': Dc, 'Dg': Dg, 'zeta': zeta}

    def calc_sigma_terms(self, only_inverse = False):
        """
        Method to compute the inverse of sigma and itnp.log determinant based on the spectral decomposition
         of the kernel matrices for all factors
        """
        for k in range(self.K):
            self.calc_sigma_terms_k(k, only_inverse)

    def calc_sigma_terms_k(self, k, only_inverse = False):
        """
        Method to compute the inverse of sigma and itnp.log determinant based on the spectral decomposition
         of the kernel matrices for a given factor k
        """
        if self.zeta[k] == 1:
            self.Sigma_inv[k, :, :] = np.eye(self.Nu)
            self.Sigma_inv_logdet[k] = 1
            self.Sigma[k, :, :] = np.eye(self.N)
        else:
            if self.kronecker:
                components = self.get_components(k)
                term1 = np.kron(components['Vg'], components['Vc'])
                term2diag = 1/ (np.repeat(components['Dg'], self.C) * np.tile(components['Dc'], self.G) + self.zeta[k] / (1-self.zeta[k]))
                term3 = np.kron(components['Vg'].transpose(), components['Vc'].transpose())
                self.Sigma_inv[k, :, :] = 1 / (1 - self.zeta[k]) * gpu_utils.dot(gpu_utils.dot(term1, np.diag(term2diag)), term3)
                self.Sigma_inv_logdet[k] = - self.Nu * np.log(1 - self.zeta[k]) + np.log(term2diag).sum()

                if not only_inverse:
                    components = self.get_components(k)
                    term1 = np.kron(components['Vg'], components['Vc'])
                    term2diag = np.repeat(components['Dg'], self.C) * np.tile(components['Dc'], self.G) + \
                                self.zeta[k] / (1 - self.zeta[k])
                    term3 = np.kron(components['Vg'].transpose(), components['Vc'].transpose())
                    self.Sigma[k, :, :] = (1 - self.zeta[k]) * gpu_utils.dot(term1,
                                                                        gpu_utils.dot(np.diag(term2diag),
                                                                                      term3))
            else:
                if self.model_groups:
                    Sigma = (1 - self.zeta[k]) * self.Kc.Kmat[self.Kc.get_best_lidx(k), self.covidx,:][:, self.covidx] * self.Kg.Kmat[k,self.groupsidx,:][:,self.groupsidx] + self.zeta[k] * np.eye(self.N)
                else:
                    Sigma = (1 - self.zeta[k]) * self.Kc.Kmat[self.Kc.get_best_lidx(k), :, :] + self.zeta[k] * np.eye(self.N)
                self.Sigma_inv[k, :, :] = np.linalg.inv(Sigma)
                self.Sigma_inv_logdet[k] = np.linalg.slogdet(self.Sigma_inv[k, :, :])[1]
                if not only_inverse:
                    self.Sigma[k, :, :] = Sigma



    def getInverseTerms_k(self, k):
        """
         Method to fetch ELBO-optimal inverse covariance matrix and its determinant for a given factor k
        """
        return {'inv': self.Sigma_inv[k, :, :], 'inv_logdet': self.Sigma_inv_logdet[k]}

    def getInverseTerms(self):
        """
        Method to fetch ELBO-optimal inverse covariance matrix and its determinant for all factors
        """
        return {'inv': self.Sigma_inv, 'inv_logdet': self.Sigma_inv_logdet}

    def get_ls(self):
        """
        Method to fetch ELBO-optimal length-scales
        """
        if hasattr(self, 'Kc') and self.Kc is not None:
            ls = self.Kc.get_ls()
        else:
            ls = np.array([np.nan] * self.K)
        return ls

    def get_x(self):
        """
        Method to get low rank group covariance matrix
        """
        x = self.Kg.get_x()
        return x

    def get_sharedness(self):
        Kgs = self.Kg.Kmat
        sharedness = np.array([np.tril(np.abs((Kgs[k,:,:] - np.eye(self.G)))).sum()/(self.G * (self.G -1)/2) for k in range(self.K)])
        return sharedness

    def get_sigma(self):
        """
        Method to get sigma hyperparameter
        """
        sigma = self.Kg.get_sigma()
        return sigma

    def getExpectation(self):
        """
        Method to get sigma hyperparameter
        """
        return self.Sigma

    def getExpectations(self):
        """
        Method to get covariance matrix and inverse terms (used for sparse GPs)
        """
        Sigma = self.getExpectation()
        invTerms = self.getInverseTerms()
        return {'cov' : Sigma, 'inv': invTerms['inv'], 'inv_logdet': invTerms['inv_logdet']}

    def get_zeta(self):
        """
        Method to fetch noise parameter
        """
        return self.zeta

    def getParameters(self):
        """
        Method to fetch ELBO-optimal length-scales, improvements compared to diagonal covariance prior and structural positions
        """
        ls = self.get_ls()
        zeta = self.get_zeta()

        if not self.model_groups:
            Kg = np.ones([self.K, self.G, self.G])
            return {'l': ls, 'scale': 1 - zeta, 'sample_cov': self.sample_cov_transformed, 'Kg' : Kg}

        x = self.get_x()
        sigma = self.get_sigma()
        return {'l':ls, 'scale': 1-zeta, 'sample_cov': self.sample_cov_transformed,  'x': x, 'sigma' : sigma, 'Kg' : self.Kg.Kmat}

    def removeFactors(self, idx, axis=1):
        """
        Method to remove factors
        """
        if self.Kg is not None:
            self.Kg.removeFactors(idx)
        if self.Kc is not None:
            self.Kc.removeFactors(idx)
        self.zeta = np.delete(self.zeta, axis=0, obj=idx)
        self.updateDim(0, self.dim[0] - len(idx))
        self.K = self.K - 1
        self.Sigma  = np.delete(self.Sigma, axis=0, obj=idx)
        self.Sigma_inv  = np.delete(self.Sigma_inv, axis=0, obj=idx)
        self.Sigma_inv_logdet  = np.delete(self.Sigma_inv_logdet, axis=0, obj=idx)

    def calc_neg_elbo_k(self, par, lidx, k, var):

        self.zeta[k] = par[0]
        self.Kc.set_gridix(lidx, k)

        # if required set group parameters
        if self.model_groups:
            sigma = par[1]
            x = par[2:]

            assert len(x) == self.Kg.rank * self.G,\
                "Length of x incorrect: Is %s, should be  %s * %s" % (len(x), self.Kg.rank, self.G)
            x = x.reshape(self.Kg.rank, self.G)
            self.Kg.set_parameters(x=x, sigma=sigma, k=k, spectral_decomp=self.kronecker) # set and recalculate group kernel (matrix and spectral decomposition if Kronecker

        self.calc_sigma_terms_k(k, only_inverse = True)
        elbo = var.calculateELBO_k(k)

        return -elbo

    def calc_neg_elbo_grad_k(self, par, lidx, k, var):
        gradient_Sigma_zeta, gradient_Sigma_sigma, gradient_Sigma_x = self.calc_gradient_Sigma(par, lidx, k)

        gradient = [-var.calcELBOgrad_k(k, gradient_Sigma_zeta)] + \
                   [-var.calcELBOgrad_k(k, gradient_Sigma_sigma)] +\
                   [-var.calcELBOgrad_k(k, gradient_Sigma_x[i]) for i in range(len(gradient_Sigma_x))]

        return gradient

    def calc_Sigma_element(self, par, lidx, k, id1, id2):
        """
        Method to calculated elements of Sigma matrix in hyperparameters
        Only used for debugging purposes
        """
        self.zeta[k] = par[0]
        # set lengthscale parameter
        self.Kc.set_gridix(lidx, k)

        # if required set group parameters
        if self.model_groups:
            sigma = par[1]
            x = par[2:]

            assert len(x) == self.Kg.rank * self.G, \
                "Length of x incorrect: Is %s, should be  %s * %s" % (len(x), self.Kg.rank, self.G)
            x = x.reshape(self.Kg.rank, self.G)
            self.Kg.set_parameters(x=x, sigma=sigma, k=k,
                                   spectral_decomp=self.kronecker)

        if self.kronecker:
            Vc, Dc = self.Kc.get_kernel_components_k(k)
            Kc = gpu_utils.dot(gpu_utils.dot(Vc, np.diag(Dc)), Vc.transpose())
            Vg, Dg = self.Kg.get_kernel_components_k(k)
            Kg = gpu_utils.dot(gpu_utils.dot(Vg, np.diag(Dg)), Vg.transpose())
            val = (1-self.zeta[k]) * np.kron(Kg, Kc) + self.zeta[k] * np.eye(self.Nu)
        else:
            Kc = self.Kc.Kmat[self.Kc.get_best_lidx(k),:,:]
            Kg = self.Kg.Kmat[k,:,:]
            if self.model_groups:
                val = (1-self.zeta[k]) * Kc[ self.covidx, :][:,self.covidx] * Kg[self.groupsidx, :][:, self.groupsidx] + self.zeta[k] *np.eye(self.Nu)
            else:
                val = (1-self.zeta[k]) * Kc + self.zeta[k] *np.eye(self.Nu)

        return val[id1,id2]

    def calc_gradient_Sigma(self, par, lidx, k):
        """
        Method to calculate gradients of covariance matrix wrt to hyperparameters
        """
        self.zeta[k] = par[0]
        self.Kc.set_gridix(lidx, k)

        # if required set group parameters
        if self.model_groups:
            sigma = par[1]
            x = par[2:]

            assert len(x) == self.Kg.rank * self.G, \
                "Length of x incorrect: Is %s, should be  %s * %s" % (len(x), self.Kg.rank, self.G)
            x = x.reshape(self.Kg.rank, self.G)
            self.Kg.set_parameters(x=x, sigma=sigma, k=k,
                                   spectral_decomp=self.kronecker)  # set and recalculate group kernel (matrix and spectral decomposition if Kronecker

        # get kernel matrices
        if self.kronecker: # TODO avoid building the full matrix use V and D below
            Vc, Dc = self.Kc.get_kernel_components_k(k)
            Kc = gpu_utils.dot(gpu_utils.dot(Vc, np.diag(Dc)), Vc.transpose())
            Vg, Dg = self.Kg.get_kernel_components_k(k)
            Kg = gpu_utils.dot(gpu_utils.dot(Vg, np.diag(Dg)), Vg.transpose())

            # gradient wrt zeta
            gradient_Sigma_zeta = - np.kron(Kg, Kc) + np.eye(self.Nu)

        else:
            Kc = self.Kc.Kmat[self.Kc.get_best_lidx(k),:,:]
            Kg = self.Kg.Kmat[k,:,:]

            # gradient wrt zeta
            if self.model_groups:
                gradient_Sigma_zeta = - Kc[self.covidx, :][:,self.covidx] *\
                                      Kg[self.groupsidx, :][:, self.groupsidx] +\
                                      np.eye(self.Nu)
            else:
                gradient_Sigma_zeta = - Kc + np.eye(self.Nu)

        if self.model_groups:
            # gradient wrt sigma
            Z = np.dot(x.transpose(), x) # diagonal can be neglected as set to 1, gradient 0
            Gmat_unscaled = Z + sigma * np.eye(self.G) # this is Kg before scaled to correlation
            Gmat_unscaled_sqrt = np.sqrt(Gmat_unscaled)
            N = np.outer(np.diag(Gmat_unscaled_sqrt), np.diag(Gmat_unscaled_sqrt))
            # N = np.array([[Gmat_unscaled_sqrt[g,g] * Gmat_unscaled_sqrt[h,h] for g in range(self.G)] for h in range(self.G)])
            tmp = -0.5 * np.outer(np.diag(Gmat_unscaled_sqrt), 1/np.diag(Gmat_unscaled_sqrt))
            AN_sigma = tmp + tmp.transpose()
            # AN_sigma = np.array([[-0.5 * Gmat_unscaled_sqrt[g,g] / Gmat_unscaled_sqrt[h,h] -0.5 * Gmat_unscaled_sqrt[h,h] / Gmat_unscaled_sqrt[g,g] for g in range(self.G)] for h in range(self.G)])
            N2 = N**2
            # N2  = np.array([[Gmat_unscaled[g,g] * Gmat_unscaled[h,h] for g in range(self.G)]for h in range(self.G)])
            # AZ_sigma = 0
            diffGmat_sigma = (1-np.eye(self.G)) * Z * AN_sigma / N2
            if self.kronecker:
                gradient_Sigma_sigma = (1 - self.zeta[k]) * np.kron(diffGmat_sigma, Kc)
            else:
                gradient_Sigma_sigma = (1 - self.zeta[k]) *\
                                       diffGmat_sigma[self.groupsidx, :][:,self.groupsidx] \
                                       * Kc[self.covidx, :][:, self.covidx]

            # gradient wrt x
            gradient_Sigma_x = []
            for r in range(self.Kg.rank):
                drg = - 1/np.diag(Gmat_unscaled_sqrt) * x[r,:]
                for g in range(self.G):
                    tmp = np.outer(np.diag(Gmat_unscaled_sqrt), drg[g] * np.eye(self.G)[g, :])
                    AN_x = tmp + tmp.transpose()
                    tmp = np.outer(x[r, :], np.eye(self.G)[g, :])
                    AZ_x = tmp + tmp.transpose()
                    diffGmat_x = (1-np.eye(self.G)) * (Z * AN_x + AZ_x * N) / N2
                    if self.kronecker:
                        grad = (1 - self.zeta[k]) * np.kron(diffGmat_x, Kc)
                    else:
                        grad = (1 - self.zeta[k]) * diffGmat_x[self.groupsidx, :][:,self.groupsidx] *  Kc[self.covidx, :][:,self.covidx]
                    gradient_Sigma_x.append(grad)


            # drg = [[-0.5 * 1/ Gmat_unscaled_sqrt[g,g] * 2 * x[r, g] for r in range(self.Kg.rank)] for g in range(self.G)]
            # # below diagonal can be neglected as set to 1, gradient 0
            # AN_x = [[np.outer(np.diag(Gmat_unscaled_sqrt), drg[g][r] * np.eye(self.G)[g, :]) + np.outer(np.diag(Gmat_unscaled_sqrt), drg[g][r] * np.eye(self.G)[g, :]).transpose() for r in range(self.Kg.rank)] for g in range(self.G)]
            # AZ_x = [[np.outer(x[r, :], np.eye(self.G)[g, :]) + np.outer(x[r, :],np.eye(self.G)[g,:]).transpose() for r in range(self.Kg.rank)] for g in range(self.G)]
            # diffGmat_x  = [[(1-np.eye(self.G)) * (Z * AN_x[g][r] + AZ_x[g][r] * N) / N2 for r in range(self.Kg.rank)] for g in range(self.G)]
            # gradient_Sigma_x = [(1 - self.zeta[k]) *
            #                     diffGmat_x[g][r][self.groupsidx, :][:,self.groupsidx] *
            #                     Kc[self.covidx, :][:,self.covidx]
            #                     for r in range(self.Kg.rank) for g in range(self.G)]
        else:
            gradient_Sigma_sigma = None
            gradient_Sigma_x = None

        return gradient_Sigma_zeta, gradient_Sigma_sigma, gradient_Sigma_x

    def check_gradient(self, par_init, lidx, k, var):
        """
        Method to check analytical gradient were correctly calculated
        (Used for debugging purposes only)
        """
        # check gradients are calulcated correctly
        z = np.random.uniform(0, 1, len(par_init))
        s.optimize.approx_fprime(z, self.calc_neg_elbo_k, 1.4901161193847656e-08, lidx, k, var) - self.calc_neg_elbo_grad_k(z, lidx, k, var)
        a = np.zeros(len(par_init))
        for n in range(3):
            z = np.random.uniform(0, 1, len(par_init))
            gradient_Sigma_zeta, gradient_Sigma_sigma, gradient_Sigma_x = self.calc_gradient_Sigma(z, lidx, k)
            G_sigma_calc = [gradient_Sigma_zeta ]+ [gradient_Sigma_sigma] + gradient_Sigma_x
            G_sigma_approx = np.array([[s.optimize.approx_fprime(z, self.calc_Sigma_element, 1.4901161193847656e-08, lidx, k, i, j) for i in range(self.Nu)] for j in range(self.Nu)])
            # F_calc = np.array([[ self.sigma_fun(z, lidx, k, var, i, j) for i in range(self.Nu)] for j in range(self.Nu)])
            for l in range(len(a)):
                a[l] = np.sum(np.abs(G_sigma_approx[:,:,l] - G_sigma_calc[l]))
            print("Sum of absolute differences in gradient of Sigma for lidx", lidx,":", a)
            print("Numerical ELBO gradient:", s.optimize.approx_fprime(z, self.calc_neg_elbo_k, 1.4901161193847656e-08, lidx, k, var))
            print("Analytical ELBO gradient:", self.calc_neg_elbo_grad_k(z, lidx, k, var))
            print("ELBO value:", self.calc_neg_elbo_k(z, lidx, k, var))
            print("Difference in ELBO gradient:", s.optimize.check_grad(self.calc_neg_elbo_k, self.calc_neg_elbo_grad_k, z, lidx, k, var))

    def optimise(self, var):
        """
        Method to find for each factor the lengthscale parameter and remaining hyperparameters that optimises the ELBO of the factor.
        The optimization can be carried out on a per-factor basis (not required on all combinations) as latent variables are independent in the elbo

        Parameters:
            - var: Node to interact with
        """
        print('Optimising sigma node...')

        K = var.dim[1]
        assert K == len(self.zeta) and K == self.K, \
            'problem in dropping factor'

        # optimise hyperparamters of GP
        for k in range(K):

            best_zeta = -1
            best_elbo = -np.Inf

            # set initial values for optimization
            if self.model_groups:
                par_init = np.hstack([self.zeta[k], self.get_sigma()[k].flatten(), self.get_x()[k, :, :].flatten()])
            else:
                par_init = self.zeta[k]

            # use grid search to optimise lengthscale hyperparameters
            for lidx in range(self.n_grid):
                if self.model_groups:
                    bounds = [(1e-10, 1 - 1e-10)]  # zeta
                    bounds = bounds + [(1e-10, 1 - 1e-10)]  # sigma
                    bounds = bounds + [(-1, 1)] * self.G * self.Kg.rank  # x
                    par0 = [1, 1] + [0] * self.G * self.Kg.rank  # parameters for zero lengthscale/scale (unstrucutred)
                    # par0 = [1, 0] + [1/np.sqrt(self.Kg.rank)] * self.G *  self.Kg.rank # parameters for zero lengthscale (fully connected groups)
                else:
                    bounds = [(1e-10, 1 - 1e-10)]  # zeta
                    par0 = [1]
                par0 = np.array(par0)

                # make sure initial parameters are in admissible region
                par_init = np.max(np.vstack([par_init, [bounds[k][0] for k in range(len(bounds))]]), axis=0)
                par_init = np.min(np.vstack([par_init, [bounds[k][1] for k in range(len(bounds))]]), axis=0)

                # optimize
                if self.use_gradients:
                    # self.check_gradient(par_init, lidx, k, var) # for debugging purposes
                    res = s.optimize.minimize(self.calc_neg_elbo_k, args=(lidx, k, var), x0=par_init, bounds=bounds,
                                              jac=self.calc_neg_elbo_grad_k)  # L-BFGS-B
                else:
                    res = s.optimize.minimize(self.calc_neg_elbo_k, args=(lidx, k, var), x0=par_init, bounds=bounds)
                elbo = -res.fun

                # for lidx = 0: zeta can be non-identifiable, use unstructured model (do not model soley group structure)
                if lidx == 0:
                    best_param4ls = par0
                else:
                    best_param4ls = res.x

                # for zeta = 1: K_G and K_C are non-identifiable, use unstructured model
                if best_param4ls[0] == 1:
                    best_param4ls = par0
                    lidx = 0

                if elbo > best_elbo:
                    best_elbo = elbo
                    best_lidx = lidx
                    best_zeta = best_param4ls[0]

                    if self.model_groups:
                        best_sigma = best_param4ls[1]
                        best_x = best_param4ls[2:].reshape(self.Kg.rank, self.G)

            # save optimized kernel paramters
            self.Kc.set_gridix(best_lidx, k)
            if self.model_groups:
                self.Kg.set_parameters(x=best_x, sigma=best_sigma, k=k, spectral_decomp=self.kronecker)

            self.zeta[k] = best_zeta

        self.calc_sigma_terms(only_inverse=False)
        # print('Sigma node has been optimised:\n- Lengthscales = %s \n- Scale = %s' % \
        #     (np.array2string(self.get_ls(), precision=2, separator=", "), np.array2string(1-self.get_zeta(), precision=2, separator=", ")))

    def updateParameters(self, ix, ro):
        """
        Public method to update the nodes parameters
        """
        self.iter += 1
        if self.iter >= self.start_opt and self.iter % self.opt_freq == 0:
            var = self.markov_blanket['Z']
            self.optimise(var)

    def calculateELBO(self): # no contribution to ELBO
        return 0

    def get_mini_batch(self):
        """
        Method to fetch minibatch
        """
        if self.mini_batch is None:
            return self.getExpectations()
        else:
            return self.mini_batch


class Sigma_Node(Sigma_Node_base):

    def __init__(self, dim, sample_cov, groups, start_opt=20, opt_freq = 10,
                 n_grid = 10, rankx = None,  model_groups = False):

        super().__init__(dim, sample_cov, groups, start_opt, opt_freq, n_grid, rankx, model_groups)

        # initialize group kernel
        if self.model_groups:
            self.G = len(self.groups)  # number of groups
            if rankx is None:
                rankx = 1
            self.kronecker = np.all([np.all(self.sample_cov_transformed[self.groupsidx == 0] ==
                                             self.sample_cov_transformed[self.groupsidx == g]) for g in
                                             range(self.G)])
            self.initKg(rank = rankx, spectral_decomp = self.kronecker)
        else:
            # all samples are modelled jointly in the covariate kernel
            self.Kg = None
            self.kronecker = True
            self.G = 1

        # initialize Sigma terms (unstructured)
        self.Nu = self.N                                            # dimension of the inverse matrix (can differ for sparse subclasses)
        self.Sigma_inv = np.zeros([self.K, self.Nu, self.Nu])
        self.Sigma = np.zeros([self.K, self.N, self.N])
        for k in range(self.K):
            self.Sigma_inv[k, :, :] = np.eye(self.Nu)
            self.Sigma[k, :, :] = np.eye(self.N)
        self.Sigma_inv_logdet = np.zeros(self.K)

        # initialize covariate kernel
        self.initKc(self.sample_cov_transformed, spectral_decomp=self.kronecker)



class Sigma_Node_sparse(Sigma_Node_base):
    """
        Sigma node making use of sparse inference.

       Additional parameters:
        - idx_inducing: Index of inducing points (default None - models the full kernel matrix)
    """

    def __init__(self, dim, sample_cov, groups, start_opt=20, opt_freq = 10,
                 n_grid=10, rankx = None, model_groups = False, idx_inducing = None):

        super().__init__(dim, sample_cov, groups, start_opt, opt_freq, n_grid, rankx, model_groups)

        # sparse GPs
        self.idx_inducing = idx_inducing
        self.Nu = len(idx_inducing)
        self.groupsidx_all = self.groupsidx
        self.groupsidx = self.groupsidx[self.idx_inducing]                  # subset to group labels for inducing points

        # initialize group kernel
        if self.model_groups:
            self.G = len(self.groups)  # number of groups
            if rankx is None:
                rankx = 1
            self.kronecker = np.all([np.all(self.sample_cov_transformed[self.idx_inducing][self.groupsidx == 0] ==
                                             self.sample_cov_transformed[self.idx_inducing][self.groupsidx == g]) for g in
                                             range(self.G)])
            self.initKg(rank = rankx, spectral_decomp = self.kronecker)
        else:
            # all samples are modelled jointly in the covariate kernel
            self.Kg = None
            self.kronecker = True
            self.G = 1


        # initialize Sigma terms (unstructured)
        self.Sigma_inv = np.zeros([self.K, self.Nu, self.Nu])
        self.Sigma = np.zeros([self.K, self.N, self.N])
        for k in range(self.K):
            self.Sigma_inv[k, :, :] = np.eye(self.Nu)
            self.Sigma[k, :, :] = np.eye(self.N)
        self.Sigma_inv_logdet = np.zeros(self.K)


        # initialize sparse covariance kernel
        self.initKc(self.sample_cov_transformed[self.idx_inducing], cov4grid=self.sample_cov_transformed,
                    spectral_decomp=self.kronecker)  # use all point to determine grid limits

        # exclude cases not covered
        # assert self.model_groups is False, "The option model_groups is not yet implemented in conjunction with sparse GPs"


    def calc_sigma_terms_k(self, k, only_inverse = False):
        """
        Method to compute the inverse of sigma and itnp.log determinant based on the spectral decomposition
         of the kernel matrices for a given factor k
        """
        if self.zeta[k] == 1:
            self.Sigma_inv[k, :, :] = np.eye(self.Nu)
            self.Sigma_inv_logdet[k] = 1
            self.Sigma[k, :, :] = np.eye(self.N)
        else:
            if self.kronecker:
                super().calc_sigma_terms_k(k, only_inverse = True)
                if not only_inverse:
                    self.update_Sigma_complete_k(k)

            else:
                    self.update_Sigma_complete_k(k)
                    # TODO make use of Kronecker structures where possible in the following
                    self.Sigma_inv[k, :, :] = np.linalg.inv(self.Sigma[k, self.idx_inducing, :][:,self.idx_inducing])
                    self.Sigma_inv_logdet[k] = np.linalg.slogdet(self.Sigma_inv[k, :, :])[1]



    def update_Sigma_complete_k(self, k):
        """
        Method to update the entire covariance matrix for the k-th factor (used in context of sparse GPs to pass to Z|U node)
        """
        if self.zeta[k] == 1 or self.Kc is None:
                self.Sigma[k, :, :] = np.eye(self.N)
        else:
            Kc_k = self.Kc.eval_at_newpoints_k(self.sample_cov_transformed, k)
            if not self.model_groups:
                self.Sigma[k, :, :] = (1 - self.zeta[k]) * Kc_k  + self.zeta[k] * np.eye(self.N)
            else:
                self.Sigma[k, :, :] = (1 - self.zeta[k]) *\
                                      Kc_k * self.Kg.Kmat[k, self.groupsidx_all, :][:, self.groupsidx_all] + \
                                      self.zeta[k] * np.eye(self.N)

    def updateParameters(self, ix, ro):
        """
        Public method to update the nodes parameters
        """
        self.iter += 1
        if self.iter >= self.start_opt and self.iter % self.opt_freq == 0:
            var = self.markov_blanket['U']
            self.optimise(var)



class Sigma_Node_warping(Sigma_Node_base):
    """
        Sigma node with the ability to align covariates acorss sample groups (for one-dimensional covariates)

       Additional parameters (compared to Sigma_Node)
        - warping: Boolean, whether to perform warping of covariates across groups in the latent space
        - warping_freq: how often to perform warping, every n-th iteration
        - warping_ref: reference group for warping
        - warping_open_begin: Allow free beginning for the warped covariates in each group?
        - warping_open_end:  Allow free end for the warped covariates in each group?
    """

    def __init__(self, dim, sample_cov, groups, start_opt=20, opt_freq = 10,
                 n_grid=10, rankx = None, model_groups = False,
                 warping_freq = 20, warping_ref = 0, warping_open_begin = True,
                 warping_open_end = True, warping_groups = None):

        super().__init__(dim, sample_cov, groups, start_opt, opt_freq, n_grid, rankx, model_groups)
        self.kronecker = False
        self.new_alignment = False

        # initialize group kernel
        if self.model_groups:
            self.G = len(self.groups)  # number of groups
            if rankx is None:
                rankx = 1
            self.initKg(rank = rankx, spectral_decomp = self.kronecker)
        else:
            # all samples are modelled jointly in the covariate kernel
            self.Kg = None
            self.G = 1

        # set warping options
        self.G4warping =  len(pd.factorize(warping_groups)[1]) # number of groups to consider for warping (if no group kernel this differs from self.G)
        assert warping_ref < self.G4warping,\
            "Reference group not correctly specified, exceeds the number of groups."
        self.reference_group = warping_ref
        self.warping_freq = warping_freq
        self.warping_open_begin = warping_open_begin
        self.warping_open_end = warping_open_end
        self.warping_groups = pd.factorize(warping_groups)[0]

        assert self.start_opt % self.warping_freq == 0,\
            "start_opt should be a multiple of warping_freq"            # to ensure in the first opt. step alignment is performed

        # initialize Sigma terms (unstructured)
        self.Nu = self.N                                            # dimension of the inverse matrix (can differ for sparse subclasses)
        self.Sigma_inv = np.zeros([self.K, self.Nu, self.Nu])
        self.Sigma = np.zeros([self.K, self.N, self.N])
        for k in range(self.K):
            self.Sigma_inv[k, :, :] = np.eye(self.Nu)
            self.Sigma[k, :, :] = np.eye(self.N)
        self.Sigma_inv_logdet = np.zeros(self.K)

        # covariate kernel is initialized after first warping
        self.Kc = None


    def updateParameters(self, ix, ro):
        """
        Public method to update the nodes parameters
        """
        self.iter += 1
        if self.iter >= self.start_opt and self.iter % self.warping_freq == 0:
            ZE = self.markov_blanket['Z'].getExpectation()
            self.align_sample_cov_dtw(ZE)
            print("Covariates were aligned between groups.")
            self.new_alignment = True

        if self.iter >= self.start_opt and self.iter % self.opt_freq == 0:
            var = self.markov_blanket['Z']
            self.optimise(var)

            # after a new alignment check that a sufficient smoothness and sharedness is present for a sensible alignment
            if self.new_alignment:
                if self.model_groups:
                    if (all(1-self.get_zeta() < 0.1) or all(self.get_sharedness()< 0.1)):
                        print("WARNING: Factors show little shared smooth variation between groups - alignment might not work as intended.")
                else:
                    if all(1 - self.get_zeta() < 0.1):
                        print("WARNING: Factors show little smooth variation - alignment might not work as intended.")
                self.new_alignment = False


    def align_sample_cov_dtw(self, Z):
        """
        Method to perform DTW between groups in the factor space.
        """
        try:
            from dtw import dtw # note this is dtw-python not dtw
        except ImportError:
            print("dtw-python module not found. This is required for alignment.")
            pass

        paths = []
        for g in range(self.G4warping):
            if g is not self.reference_group:
                tg = np.sort(np.unique(self.sample_cov[self.warping_groups == g,0]))
                Zg = [np.mean(Z[self.warping_groups == g, :][self.sample_cov[self.warping_groups == g,0] == t,:], axis =0) for t in tg]
                tref = np.sort(np.unique(self.sample_cov[self.warping_groups == self.reference_group,0]))
                Zref= [np.mean(Z[self.warping_groups == self.reference_group, :][self.sample_cov[self.warping_groups == self.reference_group,0] == t,:], axis =0) for t in tref]
                step_pattern = "asymmetric" # matches each element in the query to exactly one in the reference
                # step_pattern = "asymmetric" if self.warping_open_begin or self.warping_open_end else "symmetric2"
                alignment = dtw(np.array(Zg),
                                np.array(Zref),
                                open_begin=self.warping_open_begin, open_end=self.warping_open_end,
                                step_pattern=step_pattern)
                ref_idx = alignment.index2
                new_val = tref[ref_idx]
                old_val = self.sample_cov[self.warping_groups == g, 0]
                new_sample_cov = [new_val[tg == told].item() for told in old_val]
                self.sample_cov_transformed[self.warping_groups == g, 0] = new_sample_cov

                # # reorder by covariate value to ensure monotonicity constrains are correctly placed
                # idx_ref_order = np.argsort(self.sample_cov[self.warping_groups == self.reference_group,0])
                # idx_query_order = np.argsort(self.sample_cov[self.warping_groups == g,0])
                # # allow for partial matching (no corresponding end and beginning)
                # step_pattern = "asymmetric" if self.warping_open_begin or self.warping_open_end else "symmetric2"
                # alignment = dtw(Z[self.warping_groups == g, :][idx_query_order,:], Z[self.warping_groups == self.reference_group, :][idx_ref_order,:],
                #                 open_begin = self.warping_open_begin, open_end = self.warping_open_end, step_pattern=step_pattern)
                # query_idx = alignment.index1 # dtw-python
                # ref_idx = alignment.index2
                # ref_val = self.sample_cov[self.warping_groups == self.reference_group, 0][idx_ref_order][ref_idx]
                # idx = np.where(self.warping_groups == g)[0][idx_query_order][query_idx]
                # self.sample_cov_transformed[idx, 0] = ref_val

        # covariate kernel needs to be re-initialized after each warping
        self.initKc(self.sample_cov_transformed, spectral_decomp=self.kronecker)

#
# class Sigma_Node_warping_sparse(Sigma_Node_warping, Sigma_Node_sparse):
#     """
#         Sparse Sigma node with the ability to align covariates across sample groups (for one-dimensional covariates)
#     """
#     def __init__(self, dim, sample_cov, groups, start_opt=20, opt_freq = 10, rankx = None,
#                  model_groups = False, idx_inducing = None, warping_freq = 20, warping_ref = 0,
#                  warping_open_begin = True, warping_open_end = True):
#
#         # TODO check this works as intended
#         super().__init__(dim, sample_cov, groups, start_opt, opt_freq, rankx, model_groups, idx_inducing,
#                          warping_freq, warping_ref, warping_open_begin, warping_open_end)
#
#     def updateParameters(self, ix, ro):
#         """
#         Public method to update the nodes parameters
#         """
#         self.iter += 1
#
#         if self.iter % self.warping_freq == 0:
#             ZE = self.markov_blanket['U'].markov_blanket['Z'].getExpectation()  # use all factor values for alignment
#             self.align_sample_cov_dtw(ZE)
#             print("Covariates were aligned between groups.")
#
#         if self.iter >= self.start_opt and self.iter % self.opt_freq == 0:
#             var = self.markov_blanket['Z']
#             self.optimise(var)


# # gpytorch based implementation of the above - expects data with Kronecker structure
# class Sigma_Node_torch(Sigma_Node):
#     def __init__(self, dim, sample_cov, groups, start_opt=20, n_grid=10, idx_inducing=None,
#                  warping=False, warping_freq=20, warping_ref=0, warping_open_begin=True,
#                  warping_open_end=True, opt_freq=10, rankx=None, sigma_const=True,
#                  model_groups=False, torch_seed = 7823982, gp_iter = 200, verbose = False):
#         super().__init__(dim, sample_cov, groups, start_opt, n_grid, idx_inducing,
#                  warping, warping_freq, warping_ref, warping_open_begin,
#                  warping_open_end, opt_freq, rankx, sigma_const,
#                  model_groups)
#
#         self.gp = [None] * self.K
#         self.likelihood = [myMultitaskGaussianLikelihood(num_tasks=self.G, noise_constraint=gpytorch.constraints.Interval(1e-4,1 - 1e-4), rank = 0)] * self.K # noise constraint from original mutltiakslik, no correlation model for noise (rank = 0),  ELBO instead of MLL
#         # self.likelihood = [MultitaskGaussianLikelihood(num_tasks=self.G, rank = 0)] * self.K # basic multitaks model (no ELBO term and noise/scale dependence)
#
#         self.l_limits = gp_utils.get_l_limits(sample_cov)
#         self.sigma = np.array([np.nan] * self.K)
#         self.ls = np.array([np.nan] * self.K)
#         self.Gmat = [np.nan] * self.K
#         self.B = [np.nan] * self.K
#         self.gp_iter = gp_iter
#         self.verbose = verbose
#         torch.manual_seed(torch_seed)
#
#         # avoid initiilisation of the following - make a more gneral superclass
#         self.rank = self.Kg.rank
#         self.Kg = None
#         self.Kc = None
#
#     # def objective(self, par, lidx, k, var):
#     #     """"
#     #     Reimplements the ELBO as implemented in Z node here use pure pytorch for optimization
#     #     """
#     #     0.5 *
#
#     def optimise(self):
#         """
#         Method to find for each factor the lengthscale parameter that optimises the ELBO of the factor.
#         The optimization can be carried out on a per-factor basis (not required on all combinations) as latent variables are independent in the elbo
#         """
#
#         # get Z/U node of Markov blanket
#         if not self.idx_inducing is None:
#             var = self.markov_blanket['U']
#         else:
#             var = self.markov_blanket['Z']
#
#         K = var.dim[1]
#         assert K == len(self.zeta) and K == self.K, \
#             'problem in dropping factor'
#
#         # perform DTW to align groups
#         if self.warping and self.G4warping > 1 and self.iter % self.warping_freq == 0:
#             if not self.idx_inducing is None:
#                 Zvar = self.markov_blanket['U'].markov_blanket['Z']  # use all factor values for alignment
#             else:
#                 Zvar = var
#
#             ZE = Zvar.getExpectation()
#             self.align_sample_cov_dtw(ZE)
#             print("Covariates were aligned between groups.")
#
#         # optimise hyperparamters of GP
#         if self.iter % self.opt_freq == 0:
#             for k in range(K):
#
#                 best_i = -1
#                 best_zeta = -1
#                 best_elbo = -np.Inf
#
#                 # set initial values for optimization
#                 ZE = copy.deepcopy(var.getExpectation())
#                 ZE = torch.as_tensor(ZE, dtype=torch.float32)
#                 ytrain = torch.stack([ZE[self.groupsidx == g, k] for g in range(self.G)])
#                 xtrain = torch.as_tensor(self.sample_cov_transformed[self.groupsidx == 0], dtype=torch.float32) # TODO only works for kroncker structure RIGHT DIMS?
#
#                 # self.gp[k] = commonMultitaskGPModel(train_x=xtrain, train_y=ytrain,
#                 #                               likelihood=self.likelihood[k],
#                 #                               n_tasks=self.G, rank=self.rank)
#
#                 self.gp[k] = MyMultitaskGPModel(train_x=xtrain, train_y=ytrain,
#                                               likelihood=self.likelihood[k],
#                                               n_tasks=self.G, rank=self.rank,
#                                               var_constraint = gpytorch.constraints.Interval(1e-10,1 - 1e-10),
#                                               covar_factor_constraint=gpytorch.constraints.Interval(-1 + 1e-10,1 - 1e-10),
#                                               lengthscale_constraint = gpytorch.constraints.Interval(self.l_limits[0], self.l_limits[1]))#gpytorch.constraints.Interval(-1 + 1e-10, 1 - 1e-10))
#
#
#                 training_iterations = self.gp_iter
#
#                 # Find optimal model hyperparameters
#                 self.gp[k].train()
#                 self.likelihood[k].train()
#
#                 # Use the adam optimizer
#                 optimizer = torch.optim.Adam([
#                     {'params': self.gp[k].parameters()},  # Includes GaussianLikelihood parameters
#                 ], lr=0.1)
#
#                 # "Loss" for GP given by ELBO
#                 ZCov = copy.deepcopy(torch.as_tensor(var.getExpectations()['cov'][k], dtype=torch.float32))
#                 elbo = ELBO(self.likelihood[k], self.gp[k], ZCov)
#                 # mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.likelihood[k], self.gp[k])
#
#                 for i in range(training_iterations):
#                     optimizer.zero_grad()
#                     output = self.gp[k](xtrain)
#                     loss = -elbo(output, ytrain)
#                     # loss = -mll(output, ytrain)
#                     loss.backward()
#                     if self.verbose:
#                         print('Iter %d/%d - Loss: %.3f' % (i + 1, training_iterations, loss.item()))
#                     optimizer.step()
#
#                 self.gp[k].eval()
#                 self.likelihood[k].eval()
#
#                 print('Sigma node for factor %s has been optimised: ' %k)
#                 self.zeta[k] = (self.gp[k].likelihood.noise).detach().numpy().item()
#                 self.ls[k] = (self.gp[k].covar_module.data_covar_module.lengthscale).detach().numpy().item()
#                 self.sigma[k] = (self.gp[k].covar_module.task_covar_module.var).detach().numpy()
#                 self.B[k] = (self.gp[k].covar_module.task_covar_module.covar_factor).detach().numpy()
#                 # self.Gmat[k] = np.dot(self.B[k], self.B[k].transpose()) + np.diag(self.sigma[k]) #equivalent to below
#                 # self.Gmat[k] = gpytorch.lazy.LazyTensor.evaluate(self.gp[k].covar_module.task_covar_module.covar_matrix).detach().numpy()
#                 self.Gmat[k] = self.gp[k].covar_module.task_covar_module.covar_matrix.detach().numpy()
#
#                 # TODO avoid these recomputations build from in gp objects
#                 K_mixed = gpytorch.lazy.LazyTensor.evaluate(self.gp[k].covar_module(xtrain)).detach().numpy() # this is Kc \odot Kg not Kg \odot Kc
#                 Kmat = np.vstack([np.hstack(K_mixed[np.arange(self.N) % self.G == g][:, np.arange(self.N) % self.G == h] for h in range(self.G)) for g in range(self.G)])
#                 # self.gp[k](xtrain).precision_matrix
#                 # self.gp[k](xtrain).covariance_matrix
#                 assert np.max(np.abs(SE(self.covariates, self.ls[k], zeta=0)[self.covidx, :][:,self.covidx] *
#                                      self.Gmat[k][self.groupsidx, :][:,self.groupsidx] - Kmat)) < 1e-4,\
#                     "bug in kernel computation"
#
#                 self.Sigma[k, :, :] = (1 - self.zeta[k]) * Kmat + self.zeta[k] * np.eye(self.N)
#                 self.Sigma_inv[k, :, :] = np.linalg.inv(self.Sigma[k, :, :]) #TODO possible to reuse terms from gp object? - see above
#                 self.Sigma_inv_logdet[k] = np.linalg.slogdet(self.Sigma_inv[k, :, :])[1]
#
#
#             print('Inferred hyperparameters: Lengthscales =', self.ls, ', noise =', self.get_zeta())
#             print('Inferred hyperparameters: B =', self.B, ', sigma =', self.sigma)
#
#     def make_predictions(self, k, test_x):
#         # Make predictions
#         with torch.no_grad(), gpytorch.settings.fast_pred_var():
#             predictions = self.likelihood[k](self.gp[k](test_x))
#             mean = predictions.mean
#             lower, upper = predictions.confidence_region()
#
#         return mean, lower, upper
#
#     def get_zeta(self):
#         """
#         Method to fetch ELBO-optimal noise
#         """
#         return self.zeta
#
#     def get_ls(self):
#         """
#         Method to fetch ELBO-optimal length-scales
#         """
#         return self.ls
#
#     def get_x(self):
#         """
#         Method to get low rank group covariance matrix
#         """
#         return self.B
#
#     def get_sigma(self):
#         """
#         Method to get sigma hyperparametr
#         """
#         return self.sigma
#
#     def getParameters(self):
#         """
#         Method to fetch ELBO-optimal length-scales, improvements compared to diagonal covariance prior and structural positions
#         """
#         ls = self.get_ls()
#         zeta = self.get_zeta()
#
#         if not self.model_groups:
#             Kg = np.ones([self.K, self.G, self.G])
#             return {'l': ls, 'scale': 1 - zeta, 'sample_cov': self.sample_cov_transformed, 'Kg' : Kg}
#
#         x = self.get_x()
#         sigma = self.get_sigma()
#         return {'l':ls, 'scale': 1-zeta, 'sample_cov': self.sample_cov_transformed,  'x': x, 'sigma' : sigma, 'Kg' : self.Gmat}



from __future__ import division
import numpy.ma as ma
import numpy as np
import scipy as s
import math

from ..utils import dotd
from .. import gpu_utils

# Import manually defined functions
from .variational_nodes import Constant_Variational_Node

class Y_Node(Constant_Variational_Node):
    def __init__(self, dim, value, groups):
        Constant_Variational_Node.__init__(self, dim, value)

        # Define groups
        assert len(groups) == dim[0]
        self.groups = groups
        self.n_groups = len(np.unique(groups))

        # Mask missing values
        self.mask = self.mask()

        self.mini_batch = None
        self.mini_mask = None

    def precompute(self, options=None):
        """ Method to precompute some terms to speed up the calculations """

        # Dimensionalities
        self.N = self.dim[0] - self.getMask().sum(axis=0)
        self.D = self.dim[1] - self.getMask().sum(axis=1)

        # GPU mode
        gpu_utils.gpu_mode = options['gpu_mode']

        # Do TauTrick to speed up ELBO computation?
        # Important: this assumes that the Tau update has been done prior to calculating elbo of Y
        self.TauTrick = True

        # Constant ELBO terms
        self.likconst = -0.5 * np.sum(self.N) * np.log(2.*np.pi)

    def mask(self):
        """ Method to mask missing observations """
        mask = np.isnan(self.value)
        self.value[mask] = 0.
        return mask

    def getMask(self, full=False):
        """ Get method for the mask """
        if full:
            return self.mask
        else:
            if self.mini_batch is None:
                return self.mask
            else:
                return self.mini_mask
            
    def define_mini_batch(self, ix):
        """ Method to define a mini-batch (only for stochastic inference) """
        self.mini_batch = self.value[ix,:]
        self.mini_mask = self.mask[ix,:]

    def get_mini_batch(self):
        """ Method to retrieve a mini-batch (only for stochastic inference) """
        if self.mini_batch is None:
            return self.getExpectation()
        else:
            return self.mini_batch

    def calculateELBO(self):
        """ Method to calculate evidence lower bound """
        mask = self.mask
        Tau = self.markov_blanket["Tau"].getExpectations(expand=False)
        elbo = self.likconst
        groups = self.markov_blanket["Tau"].groups

        if self.TauTrick: 
            tauQ_param = self.markov_blanket["Tau"].getParameters("Q")
            tauP_param = self.markov_blanket["Tau"].getParameters("P")
            for g in range(len(np.unique(groups))):
                idx = groups==g
                foo = (~mask[idx,:]).sum(axis=0)
                elbo += 0.5*(Tau["lnE"][g,:]*foo).sum() - np.dot(Tau["E"][g,:],(tauQ_param["b"][g,:] - tauP_param["b"][g,:]))

        else:
            Y = self.getExpectation()
            Wtmp = self.markov_blanket["W"].getExpectations()
            Ztmp = self.markov_blanket["Z"].getExpectations()
            W, WW = Wtmp["E"].T, Wtmp["E2"].T
            Z, ZZ = Ztmp["E"], Ztmp["E2"]

            tmp = np.square(Y) \
                + ZZ.dot(WW) \
                - np.dot(np.square(Z),np.square(W)) + np.square(Z.dot(W)) \
                - 2*Z.dot(W)*Y 
            tmp *= 0.5
            tmp[mask] = 0.
            
            for g in range(len(np.unique(groups))):
                idx = groups==g
                foo = (~mask[idx,:]).sum(axis=0)
                elbo += 0.5*(Tau["lnE"][g,:]*foo).sum() - (Tau["E"][g,:]*tmp[idx,:]).sum()
        return elbo




"""Various utility functions."""
import numpy as np
import scipy.sparse as sp
import torch

from typing import Union


def l2_reg_loss(model, scale=1e-5):
    """Get L2 loss for model weights."""
    loss = 0.0
    for w in model.get_weights():
        loss += w.pow(2.).sum()
    return loss * scale


def to_sparse_tensor(matrix: Union[sp.spmatrix, torch.Tensor],
                     cuda: bool = True,
                     ) -> Union[torch.sparse.FloatTensor, torch.cuda.sparse.FloatTensor]:
    """Convert a scipy sparse matrix to a torch sparse tensor.

    Args:
        matrix: Sparse matrix to convert.
        cuda: Whether to move the resulting tensor to GPU.

    Returns:
        sparse_tensor: Resulting sparse tensor (on CPU or on GPU).

    """
    if sp.issparse(matrix):
        coo = matrix.tocoo()
        indices = torch.LongTensor(np.vstack([coo.row, coo.col]))
        values = torch.FloatTensor(coo.data)
        shape = torch.Size(coo.shape)
        sparse_tensor = torch.sparse.FloatTensor(indices, values, shape)
    elif torch.is_tensor(matrix):
        row, col = matrix.nonzero().t()
        indices = torch.stack([row, col])
        values = matrix[row, col]
        shape = torch.Size(matrix.shape)
        sparse_tensor = torch.sparse.FloatTensor(indices, values, shape)
    else:
        raise ValueError(f"matrix must be scipy.sparse or torch.Tensor (got {type(matrix)} instead).")
    if cuda:
        sparse_tensor = sparse_tensor.cuda()
    return sparse_tensor.coalesce()


def coms_list_to_matrix(communities_list, num_nodes=None):
    """Convert a communities list of len [C] to an [N, C] communities matrix.

    Parameters
    ----------
    communities_list : list
        List of lists of nodes belonging to respective community.
    num_nodes : int, optional
        Total number of nodes. This needs to be here in case
        some nodes are not in any communities, but the resulting
        matrix must have the correct shape [num_nodes, num_coms].

    Returns
    -------
    communities_matrix : np.array, shape [num_nodes, num_coms]
        Binary matrix of community assignments.
    """
    num_coms = len(communities_list)
    if num_nodes is None:
        num_nodes = max(max(cmty) for cmty in communities_list) + 1
    communities_matrix = np.zeros([num_nodes, num_coms], dtype=np.float32)
    for cmty_idx, nodes in enumerate(communities_list):
        communities_matrix[nodes, cmty_idx] = 1
    return communities_matrix


def coms_matrix_to_list(communities_matrix):
    """Convert an [N, C] communities matrix to a communities list of len [C].

    Parameters
    ----------
    communities_matrix : np.ndarray or sp.spmatrix, shape [num_nodes, num_coms]
        Binary matrix of community assignments.

    Returns
    -------
    communities_list : list
        List of lists of nodes belonging to respective community.

    """
    num_nodes, num_coms = communities_matrix.shape
    communities_list = [[] for _ in range(num_coms)]
    nodes, communities = communities_matrix.nonzero()
    for node, cmty in zip(nodes, communities):
        communities_list[cmty].append(node)
    return communities_list


def plot_sparse_clustered_adjacency(A, num_coms, z, o, ax=None, markersize=0.25):
    import seaborn as sns
    import matplotlib.pyplot as plt
    if ax is None:
        ax = plt.gca()

    colors = sns.color_palette('hls', num_coms)
    sns.set_style('white')

    crt = 0
    for idx in np.where(np.diff(z[o]))[0].tolist() + [z.shape[0]]:
        ax.axhline(y=idx, linewidth=0.5, color='black', linestyle='--')
        ax.axvline(x=idx, linewidth=0.5, color='black', linestyle='--')
        crt = idx + 1

    ax.spy(A[o][:, o], markersize=markersize)
    ax.tick_params(axis='both', which='both', labelbottom='off', labelleft='off', labeltop='off')


def adjacency_split_naive(A, p_val, neg_mul=1, max_num_val=None):
    edges = np.column_stack(sp.tril(A).nonzero())
    num_edges = edges.shape[0]
    num_val_edges = int(num_edges * p_val)
    if max_num_val is not None:
        num_val_edges = min(num_val_edges, max_num_val)

    shuffled = np.random.permutation(num_edges)
    which_val = shuffled[:num_val_edges]
    which_train = shuffled[num_val_edges:]
    train_ones = edges[which_train]
    val_ones = edges[which_val]
    A_train = sp.coo_matrix((np.ones_like(train_ones.T[0]), (train_ones.T[0], train_ones.T[1])),
                            shape=A.shape).tocsr()
    A_train = A_train.maximum(A_train.T)

    num_nodes = A.shape[0]
    num_val_nonedges = neg_mul * num_val_edges
    candidate_zeros = np.random.choice(np.arange(num_nodes, dtype=np.int32),
                                       size=(2 * num_val_nonedges, 2), replace=True)
    cne1, cne2 = candidate_zeros[:, 0], candidate_zeros[:, 1]
    to_keep = (1 - A[cne1, cne2]).astype(np.bool).A1
    val_zeros = candidate_zeros[to_keep][:num_val_nonedges]
    if to_keep.sum() < num_val_nonedges:
        raise ValueError("Couldn't produce enough non-edges")

    return A_train, val_ones, val_zeros



import numpy as np
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import random
import scanpy as sc
import networkx as nx
import pandas as pd
import anndata

sc_color=['#7CBB5F','#368650','#A499CC','#5E4D9A','#78C2ED','#866017', '#9F987F','#E0DFED',
 '#EF7B77', '#279AD7','#F0EEF0', '#1F577B', '#A56BA7', '#E0A7C8', '#E069A6', '#941456', '#FCBC10',
 '#EAEFC5', '#01A0A7', '#75C8CC', '#F0D7BC', '#D5B26C', '#D5DA48', '#B6B812', '#9DC3C3', '#A89C92', '#FEE00C', '#FEF2A1']

red_color=['#F0C3C3','#E07370','#CB3E35','#A22E2A','#5A1713','#D3396D','#DBC3DC','#85539B','#5C2B80','#5C4694']
green_color=['#91C79D','#8FC155','#56AB56','#2D5C33','#BBCD91','#6E944A','#A5C953','#3B4A25','#010000']
orange_color=['#EFBD49','#D48F3E','#AC8A3E','#7D7237','#745228','#E1C085','#CEBC49','#EBE3A1','#6C6331','#8C9A48','#D7DE61']
blue_color=['#347862','#6BBBA0','#81C0DD','#3E8CB1','#88C8D2','#52B3AD','#265B58','#B2B0D4','#5860A7','#312C6C']

sc_color_cmap = LinearSegmentedColormap.from_list('Custom', sc_color, len(sc_color))

def pyomic_plot_set(verbosity=3,dpi=80,facecolor='white'):
    sc.settings.verbosity = verbosity             # verbosity: errors (0), warnings (1), info (2), hints (3)
    sc.settings.set_figure_params(dpi=dpi, facecolor=facecolor)
    import warnings
    warnings.simplefilter("ignore", category=UserWarning)
    warnings.simplefilter("ignore", category=FutureWarning)
    warnings.simplefilter("ignore", category=DeprecationWarning)

def ov_plot_set(verbosity=3,dpi=80,facecolor='white'):
    sc.settings.verbosity = verbosity             # verbosity: errors (0), warnings (1), info (2), hints (3)
    sc.settings.set_figure_params(dpi=dpi, facecolor=facecolor)
    import warnings
    warnings.simplefilter("ignore", category=UserWarning)
    warnings.simplefilter("ignore", category=FutureWarning)
    warnings.simplefilter("ignore", category=DeprecationWarning)

def pyomic_palette()->list:
    """
    Returns a dictionary of colors for various plots used in pyomic package.

    Returns:
        sc_color: List containing the hex codes as values.
    """ 
    return sc_color

def palette()->list:
    """
    Returns a dictionary of colors for various plots used in pyomic package.

    Returns:
        sc_color: List containing the hex codes as values.
    """ 
    return sc_color

def red_palette()->list:
    """
    Returns a dictionary of colors for various plots used in pyomic package.

    Returns:
        sc_color: List containing the hex codes as values.
    """ 
    return red_color

def green_palette()->list:
    """
    Returns a dictionary of colors for various plots used in pyomic package.

    Returns:
        sc_color: List containing the hex codes as values.
    """ 
    return green_color

def orange_palette()->list:
    """
    Returns a dictionary of colors for various plots used in pyomic package.

    Returns:
        sc_color: List containing the hex codes as values.
    """ 
    return orange_color

def blue_palette()->list:
    """
    Returns a dictionary of colors for various plots used in pyomic package.

    Returns:
        sc_color: List containing the hex codes as values.
    """ 
    return blue_color

def plot_text_set(text,text_knock=2,text_maxsize=20):
    """
    Formats the text to fit in a plot by adding line breaks.

    Parameters
    ----------
    - text : `str`
        Text to format.
    - text_knock : `int`, optional
        Number of words to skip between two line breaks, by default 2.
    - text_maxsize : `int`, optional
        Maximum length of the text before formatting, by default 20.

    Returns
    -------
    - text: `str`
        Formatted text.
    """
    #print(text)
    text_len=len(text)
    if text_len>text_maxsize:
        ty=text.split(' ')
        ty_len=len(ty)
        if ty_len%2==1:
            ty_mid=(ty_len//text_knock)+1
        else:
            ty_mid=(ty_len//text_knock)
        #print(ty_mid)

        if ty_mid==0:
            ty_mid=1

        res=''
        ty_len_max=np.max([i%ty_mid for i in range(ty_len)])
        if ty_len_max==0:
            ty_len_max=1
        for i in range(ty_len):
            #print(ty_mid,i%ty_mid,i,ty_len_max)
            if (i%ty_mid)!=ty_len_max:
                res+=ty[i]+' '
            else:
                res+='\n'+ty[i]+' '
        return res
    else:
        return text
    
def ticks_range(x,width):
    """
    Returns a list of ticks for a plot.
    
    Parameters
    ----------
    - x : `int`
        Number of ticks.
    - width : `float`
        Width of the plot.

    Returns
    -------
    - ticks: `list`
        List of ticks.
    """
    nticks=[]
    pticks=[]
    start=-(x//2)
    end=(x//2)
    for i in range(x//2):
        nticks.append(start+width)
        start+=width
        pticks.append(end-width)
        end-=width
    if x%2==0:
        ticks=nticks+pticks
    elif x%2==1:
        ticks=nticks+[0]+pticks
    return ticks

def plot_boxplot(data,hue,x_value,y_value,width=0.6,title='',
                 figsize=(6,3),palette=None,fontsize=10,
                 legend_bbox=(1, 0.55),legend_ncol=1,):
    """
    Plots a boxplot with jittered points.

    Parameters
    ----------
    - data : `pandas.DataFrame`
        Dataframe containing the data to plot.
    - hue : `str`
        Column name of the dataframe containing the hue data.
    - x_value : `str`
        Column name of the dataframe containing the x-axis data.
    - y_value : `str`
        Column name of the dataframe containing the y-axis data.
    - width : `float`, optional
        Width of the boxplot, by default 0.6.
    - title : `str`, optional
        Title of the plot, by default ''.
    - figsize : `tuple`, optional
        Size of the figure, by default (6,3).
    - palette : `list`, optional
        List of colors to use for the plot, by default None.
    - fontsize : `int`, optional
        Font size of the plot, by default 10.
    - legend_bbox : `tuple`, optional
        Bounding box of the legend, by default (1, 0.55).
    - legend_ncol : `int`, optional
        Number of columns in the legend, by default 1.

    Returns
    -------
    - fig: `matplotlib.figure.Figure`
        Figure object.
    - ax: `matplotlib.axes._subplots.AxesSubplot`
        Axes object.
    """

    #获取需要分割的数据
    hue=hue
    hue_datas=list(set(data[hue]))

    #获取箱线图的横坐标
    x=x_value
    ticks=list(set(data[x]))

    #在这个数据中，我们有6个不同的癌症，每个癌症都有2个基因（2个箱子）
    #所以我们需要得到每一个基因的6个箱线图位置，6个散点图的抖动
    plot_data1={}#字典里的每一个元素就是每一个基因的所有值
    plot_data_random1={}#字典里的每一个元素就是每一个基因的随机20个值
    plot_data_xs1={}#字典里的每一个元素就是每一个基因的20个抖动值


    #箱子的参数
    #width=0.6
    y=y_value
    for hue_data,num in zip(hue_datas,ticks_range(len(hue_datas),width)):
        data_a=[]
        data_a_random=[]
        data_a_xs=[]
        for i,k in zip(ticks,range(len(ticks))):
            test_data=data.loc[((data[x]==i)&(data[hue]==hue_data)),y].tolist()
            data_a.append(test_data)
            if len(test_data)<20:
                data_size=len(test_data)
            else:
                data_size=20
            random_data=random.sample(test_data,data_size)
            data_a_random.append(random_data)
            data_a_xs.append(np.random.normal(k*len(hue_datas)+num, 0.04, len(random_data)))
        #data_a=np.array(data_a)
        data_a_random=np.array(data_a_random)
        plot_data1[hue_data]=data_a 
        plot_data_random1[hue_data]=data_a_random
        plot_data_xs1[hue_data]=data_a_xs

    fig, ax = plt.subplots(figsize=figsize)
    #色卡
    if palette==None:
        palette=pyomic_palette()
    #palette=["#a64d79","#674ea7"]
    #绘制箱线图
    for hue_data,hue_color,num in zip(hue_datas,palette,ticks_range(len(hue_datas),width)):
        b1=ax.boxplot(plot_data1[hue_data], 
                    positions=np.array(range(len(ticks)))*len(hue_datas)+num, 
                    sym='', 
                    widths=width,)
        plt.setp(b1['boxes'], color=hue_color)
        plt.setp(b1['whiskers'], color=hue_color)
        plt.setp(b1['caps'], color=hue_color)
        plt.setp(b1['medians'], color=hue_color)

        clevels = np.linspace(0., 1., len(plot_data_random1[hue_data]))
        for x, val, clevel in zip(plot_data_xs1[hue_data], plot_data_random1[hue_data], clevels):
            plt.scatter(x, val,c=hue_color,alpha=0.4)

    #坐标轴字体
    #fontsize=10
    #修改横坐标
    ax.set_xticks(range(0, len(ticks) * len(hue_datas), len(hue_datas)), ticks,fontsize=fontsize)
    #修改纵坐标
    yticks=ax.get_yticks()
    ax.set_yticks(yticks[yticks>=0],yticks[yticks>=0],fontsize=fontsize)

    labels = hue_datas  #legend标签列表，上面的color即是颜色列表
    color = palette
    #用label和color列表生成mpatches.Patch对象，它将作为句柄来生成legend
    patches = [ mpatches.Patch(color=color[i], label="{:s}".format(labels[i]) ) for i in range(len(hue_datas)) ] 
    ax.legend(handles=patches,bbox_to_anchor=legend_bbox, ncol=legend_ncol,fontsize=fontsize)

    #设置标题
    ax.set_title(title,fontsize=fontsize+2)
    #设置spines可视化情况
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    return fig,ax

def plot_network(G:nx.Graph,G_type_dict:dict,G_color_dict:dict,pos_type:str='spring',pos_dim:int=2,
                figsize:tuple=(4,4),pos_scale:int=10,pos_k=None,pos_alpha:float=0.4,
                node_size:int=50,node_alpha:float=0.6,node_linewidths:int=1,
                plot_node=None,plot_node_num:int=20,
                label_verticalalignment:str='center_baseline',label_fontsize:int=12,
                label_fontfamily:str='Arial',label_fontweight:str='bold',label_bbox=None,
                legend_bbox:tuple=(0.7, 0.05),legend_ncol:int=3,legend_fontsize:int=12,
                legend_fontweight:str='bold'):
    """
    Plot network graph.

    Arguments:
        G: networkx graph
        G_type_dict: dict, node type dict
        G_color_dict: dict, node color dict
        pos_type: str, node position type, 'spring' or 'kamada_kawai'
        pos_dim: int, node position dimension, 2 or 3
        figsize: tuple, figure size
        pos_scale: int, node position scale
        pos_k: float, node position k
        pos_alpha: float, node position alpha
        node_size: int, node size
        node_alpha: float, node alpha
        node_linewidths: float, node linewidths
        plot_node: list, plot node list
        plot_node_num: int, plot node number
        label_verticalalignment: str, label verticalalignment
        label_fontsize: int, label fontsize
        label_fontfamily: str, label fontfamily
        label_fontweight: str, label fontweight
        label_bbox: tuple, label bbox
        legend_bbox: tuple, legend bbox
        legend_ncol: int, legend ncol
        legend_fontsize: int, legend fontsize
        legend_fontweight: str, legend fontweight

    
    """
    
    fig, ax = plt.subplots(figsize=figsize)
    if pos_type=='spring':
        pos = nx.spring_layout(G, scale=pos_scale, k=pos_k)
    elif pos_type=='kamada_kawai':
        pos=nx.kamada_kawai_layout(G,dim=pos_dim,scale=pos_scale)
    degree_dict = dict(G.degree(G.nodes()))
    
    G_color_dict=dict(zip(G.nodes,[G_color_dict[i] for i in G.nodes]))
    G_type_dict=dict(zip(G.nodes,[G_type_dict[i] for i in G.nodes]))

    nx.draw_networkx_edges(G, pos,nodelist=list(G_color_dict.keys()), alpha=pos_alpha)
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=list(G_color_dict.keys()),
        node_size=[degree_dict[v]*node_size for v in G],
        node_color=list(G_color_dict.values()),
        alpha=node_alpha,
        linewidths=node_linewidths,
    )
    if plot_node!=None:
        hub_gene=plot_node
    else:
        hub_gene=[i[0] for i in sorted(degree_dict.items(),key=lambda x: x[1],reverse=True)[:plot_node_num]]
    
    pos1=dict()
    for i in pos.keys():
        pos1[i]=np.array([-1000,-1000])
    for i in hub_gene:
        pos1[i]=pos[i]
    #label_options = {"ec": "white", "fc": "white", "alpha": 0.6}
    nx.draw_networkx_labels(
        G,pos1,verticalalignment=label_verticalalignment,
        font_size=label_fontsize,font_family=label_fontfamily,
        font_weight=label_fontweight,bbox=label_bbox,
    )

    ax.axis("off")
    
    t=pd.DataFrame(index=G_type_dict.keys())
    t['gene_type_dict']=G_type_dict
    t['gene_color_dict']=G_color_dict
    type_color_dict={}
    for i in t['gene_type_dict'].value_counts().index:
        type_color_dict[i]=t.loc[t['gene_type_dict']==i,'gene_color_dict'].values[0]
    
    patches = [ mpatches.Patch(color=type_color_dict[i], label="{:s}".format(i) ) for i in type_color_dict.keys() ] 

    plt.legend(handles=patches,bbox_to_anchor=legend_bbox, ncol=legend_ncol,fontsize=legend_fontsize)
    leg = plt.gca().get_legend() #或leg=ax.get_legend()
    ltext = leg.get_texts()
    plt.setp(ltext, fontsize=legend_fontsize,fontweight=legend_fontweight)
    
    return fig,ax

def plot_cellproportion(adata:anndata.AnnData,celltype_clusters:str,visual_clusters:str,
                       visual_li=None,visual_name:str='',figsize:tuple=(4,6),
                       ticks_fontsize:int=12,labels_fontsize:int=12,
                       legend:bool=False):
    """
    Plot cell proportion of each cell type in each visual cluster.

    Arguments:
        adata: AnnData object.
        celltype_clusters: Cell type clusters.
        visual_clusters: Visual clusters.
        visual_li: Visual cluster list.
        visual_name: Visual cluster name.
        figsize: Figure size.
        ticks_fontsize: Ticks fontsize.
        labels_fontsize: Labels fontsize.
        legend: Whether to show legend.
    
    
    """

    b=pd.DataFrame(columns=['cell_type','value','Week'])
    
    if visual_li==None:
        adata.obs[visual_clusters]=adata.obs[visual_clusters].astype('category')
        visual_li=adata.obs[visual_clusters].cat.categories
    
    for i in visual_li:
        b1=pd.DataFrame()
        test=adata.obs.loc[adata.obs[visual_clusters]==i,celltype_clusters].value_counts()
        b1['cell_type']=test.index
        b1['value']=test.values/test.sum()
        b1['Week']=i.replace('Retinoblastoma_','')
        b=pd.concat([b,b1])
    
    plt_data2=adata.obs[celltype_clusters].value_counts()
    plot_data2_color_dict=dict(zip(adata.obs[celltype_clusters].cat.categories,adata.uns['{}_colors'.format(celltype_clusters)]))
    plt_data3=adata.obs[visual_clusters].value_counts()
    plot_data3_color_dict=dict(zip([i.replace('Retinoblastoma_','') for i in adata.obs[visual_clusters].cat.categories],adata.uns['{}_colors'.format(visual_clusters)]))
    b['cell_type_color'] = b['cell_type'].map(plot_data2_color_dict)
    b['stage_color']=b['Week'].map(plot_data3_color_dict)
    
    fig, ax = plt.subplots(figsize=figsize)
    #用ax控制图片
    #sns.set_theme(style="whitegrid")
    #sns.set_theme(style="ticks")
    n=0
    all_celltype=adata.obs[celltype_clusters].cat.categories
    for i in all_celltype:
        if n==0:
            test1=b[b['cell_type']==i]
            ax.bar(x=test1['Week'],height=test1['value'],width=0.8,color=list(set(test1['cell_type_color']))[0], label=i)
            bottoms=test1['value'].values
        else:
            test2=b[b['cell_type']==i]
            ax.bar(x=test2['Week'],height=test2['value'],bottom=bottoms,width=0.8,color=list(set(test2['cell_type_color']))[0], label=i)
            test1=test2
            bottoms+=test1['value'].values
        n+=1
    if legend!=False:
        plt.legend(bbox_to_anchor=(1.05, -0.05), loc=3, borderaxespad=0,fontsize=10)
    
    plt.grid(False)
    
    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)

    # 设置左边和下边的坐标刻度为透明色
    #ax.yaxis.tick_left()
    #ax.xaxis.tick_bottom()
    #ax.xaxis.set_tick_params(color='none')
    #ax.yaxis.set_tick_params(color='none')

    # 设置左边和下边的坐标轴线为独立的线段
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))

    plt.xticks(fontsize=ticks_fontsize,rotation=90)
    plt.yticks(fontsize=ticks_fontsize)
    plt.xlabel(visual_name,fontsize=labels_fontsize)
    plt.ylabel('Cells per Stage',fontsize=labels_fontsize)
    fig.tight_layout()
    return fig,ax

def plot_embedding_celltype(adata:anndata.AnnData,figsize:tuple=(6,4),basis:str='umap',
                            celltype_key:str='major_celltype',title:str=None,
                            celltype_range:tuple=(2,9),
                            embedding_range:tuple=(3,10),
                            xlim:int=-1000)->tuple:
    """
    Plot embedding with celltype color by omicverse

    Arguments:
        adata: AnnData object  
        figsize: figure size
        basis: embedding method
        celltype_key: celltype key in adata.obs
        title: figure title
        celltype_range: celltype range to plot
        embedding_range: embedding range to plot
        xlim: x axis limit

    Returns:
        fig : figure and axis
        ax: axis
    
    """


    cell_num_pd=pd.DataFrame(adata.obs[celltype_key].value_counts())
    if '{}_colors'.format(celltype_key) in adata.uns.keys():
        cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories.tolist(),
                        adata.uns['{}_colors'.format(celltype_key)]))
    else:
        if len(adata.obs[celltype_key].cat.categories)>28:
            cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.default_102))
        else:
            cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.zeileis_28))

    if figsize==None:
        if len(adata.obs[celltype_key].cat.categories)<10:
            fig = plt.figure(figsize=(6,4))
        else:
            print('The number of cell types is too large, please set the figsize parameter')
            return
    else:
        fig = plt.figure(figsize=figsize)
    grid = plt.GridSpec(10, 10)
    ax1 = fig.add_subplot(grid[:, embedding_range[0]:embedding_range[1]])       # 占据第一行的所有列
    ax2 = fig.add_subplot(grid[celltype_range[0]:celltype_range[1], :2]) 
    # 定义子图的大小和位置
         # 占据第二行的前两列
    #ax3 = fig.add_subplot(grid[1:, 2])      # 占据第二行及以后的最后一列
    #ax4 = fig.add_subplot(grid[2, 0])       # 占据最后一行的第一列
    #ax5 = fig.add_subplot(grid[2, 1])       # 占据最后一行的第二列

    sc.pl.embedding(
        adata,
        basis=basis,
        color=[celltype_key],
        title='',
        frameon=False,
        #wspace=0.65,
        ncols=3,
        ax=ax1,
        legend_loc=False,
        show=False
    )

    for idx,cell in zip(range(cell_num_pd.shape[0]),
                        adata.obs[celltype_key].cat.categories):
        ax2.scatter(100,
                cell,c=cell_color_dict[cell],s=50)
        ax2.plot((100,cell_num_pd.loc[cell,celltype_key]),(idx,idx),
                c=cell_color_dict[cell],lw=4)
        ax2.text(100,idx+0.2,
                cell+'('+str("{:,}".format(cell_num_pd.loc[cell,celltype_key]))+')',fontsize=11)
    ax2.set_xlim(xlim,cell_num_pd.iloc[1].values[0]) 
    ax2.text(xlim,idx+1,title,fontsize=12)
    ax2.grid(False)
    ax2.spines['top'].set_visible(False)
    ax2.spines['right'].set_visible(False)
    ax2.spines['bottom'].set_visible(False)
    ax2.spines['left'].set_visible(False)
    ax2.axis('off')

    return fig,[ax1,ax2]

def gen_mpl_labels(
    adata, groupby, exclude=(), 
    basis='X_umap',ax=None, adjust_kwargs=None, text_kwargs=None
):
    """ 
    Get locations of cluster median . Borrowed from scanpy github forum.
    """
    if adjust_kwargs is None:
        adjust_kwargs = {"text_from_points": False}
    if text_kwargs is None:
        text_kwargs = {}

    medians = {}

    for g, g_idx in adata.obs.groupby(groupby).groups.items():
        if g in exclude:
            continue
        medians[g] = np.median(adata[g_idx].obsm[basis], axis=0)

    if ax is None:
        texts = [
            plt.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()
        ]
    else:
        texts = [ax.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()]
    from adjustText import adjust_text
    adjust_text(texts, **adjust_kwargs)


"""
This module is used to define classes for statistical distributions

Each 'Distribution' class can store an arbitrary number of distributions of the same type, this is specified in the 'dim' argument

A 'Distribution' class has two main types of attributes: parameters and expectations. Both have to be defined when initialising a class.
Note that in some distributions (Gaussian mainly) a parameter is equal to an expectation. However, they are stored as separate
attributes and are not always necessarily equal due to E and M steps in the VB algorithm.

TO-DO:
- we should pass parametrs and expectations to Distribution() and perform sanity checks there
- Improve initialisation of Multivariate Gaussian
- Sanity checks on setter and getter functions
"""

import scipy as s
import numpy as np

# from mofapy2.config import settings
# from mofapy2.core.utils import *  # TODO prob not necessary ?

# General class for probability distributions
class Distribution(object):
    """ General class for a statistical distribution """
    def __init__(self, dim):
        self.dim = dim

    def density(self):
        """ General method to calculate density """
        pass
    def loglik(self):
        """ General method to calculate log likelihood """
        pass
    def sample(self):
        """ General method to sample from the distribution """
        pass
    def entropy(self):
        """ General method to calculate entropy """
        pass
    def updateExpectations(self):
        """ General method to update expectations """
        pass

    def getParameters(self):
        """ General getter function for parameters """
        return self.params

    def setParameters(self,**params):
        """ General setter function for parameters """
        self.params = params

    def getExpectation(self):
        """ General getter function for expectations """
        return self.expectations['E']

    def getExpectations(self):
        """ General setter function for expectations """
        return self.expectations

    def CheckDimensionalities(self):
        """ General method to do a sanity check on the dimensionalities """
        # p_dim = set(map(s.shape, self.params.values()))
        e_dim = set(map(s.shape, self.expectations.values()))
        # assert len(p_dim) == 1, "Parameters have different dimensionalities"
        assert len(e_dim) == 1, "Expectations have different dimensionalities"
        # assert e_dim == p_dim, "Parameters and Expectations have different dimensionality"

    def to_float32(self):
        """ Convert numpy arrays from float64 to float32 """
        for i in self.params.keys(): self.params[i] = self.params[i].astype(np.float32)
        for i in self.expectations.keys(): self.expectations[i] = self.expectations[i].astype(np.float32)


    def removeDimensions(self, axis, idx):
        """ General method to remove undesired dimensions

        PARAMETERS
        ----------
        axis: int
            axis from where to remove the elements
        idx: list or numpy array
            indices of the elements to remove
        """
        assert axis <= len(self.dim)
        assert np.all(idx < self.dim[axis])
        for k in self.params.keys(): self.params[k] = np.delete(self.params[k], idx, axis)
        for k in self.expectations.keys(): self.expectations[k] = np.delete(self.expectations[k], idx, axis)
        self.updateDim(axis=axis, new_dim=self.dim[axis]-len(idx))

    def updateDim(self, axis, new_dim):
        """ Method to update the dimensionality of a particular axis. 
        This method is a bit inefficient but we store dimensionalities with tuples and they cannot be modified

        PARAMETERS
        ----------
        axis: int
            axis to be updated
        new_dim: int
            updated dimensionality
        """
        dim = list(self.dim)
        dim[axis] = new_dim
        self.dim = tuple(dim)
